@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2011 Jeffrey Stedfast
+ *  Copyright (C) 2000-2014 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -26,6 +26,7 @@
 #ifdef __APPLE__
 #undef HAVE_POLL_H
 #undef HAVE_POLL
+typedef unsigned int nfds_t;
 #endif
 
 #include <stdio.h>
@@ -47,12 +48,14 @@
 #endif
 
 #include "gmime-gpg-context.h"
+#ifdef ENABLE_CRYPTOGRAPHY
 #include "gmime-filter-charset.h"
 #include "gmime-stream-filter.h"
 #include "gmime-stream-pipe.h"
 #include "gmime-stream-mem.h"
 #include "gmime-stream-fs.h"
 #include "gmime-charset.h"
+#endif /* ENABLE_CRYPTOGRAPHY */
 #include "gmime-error.h"
 
 #ifdef ENABLE_DEBUG
@@ -83,6 +86,11 @@ static GMimeDigestAlgo gpg_digest_id (GMimeCryptoContext *ctx, const char *name)
 
 static const char *gpg_digest_name (GMimeCryptoContext *ctx, GMimeDigestAlgo digest);
 
+static gboolean gpg_get_retrieve_session_key (GMimeCryptoContext *context);
+
+static int gpg_set_retrieve_session_key (GMimeCryptoContext *ctx, gboolean retrieve_session_key,
+					 GError **err);
+
 static int gpg_sign (GMimeCryptoContext *ctx, const char *userid,
 		     GMimeDigestAlgo digest, GMimeStream *istream,
 		     GMimeStream *ostream, GError **err);
@@ -104,6 +112,10 @@ static int gpg_encrypt (GMimeCryptoContext *ctx, gboolean sign, const char *user
 static GMimeDecryptResult *gpg_decrypt (GMimeCryptoContext *ctx, GMimeStream *istream,
 					GMimeStream *ostream, GError **err);
 
+static GMimeDecryptResult *gpg_decrypt_session (GMimeCryptoContext *ctx, const char *session_key,
+						GMimeStream *istream, GMimeStream *ostream,
+						GError **err);
+
 static int gpg_import_keys (GMimeCryptoContext *ctx, GMimeStream *istream,
 			    GError **err);
 
@@ -155,18 +167,20 @@ g_mime_gpg_context_class_init (GMimeGpgContextClass *klass)
 	crypto_class->verify = gpg_verify;
 	crypto_class->encrypt = gpg_encrypt;
 	crypto_class->decrypt = gpg_decrypt;
+	crypto_class->decrypt_session = gpg_decrypt_session;
 	crypto_class->import_keys = gpg_import_keys;
 	crypto_class->export_keys = gpg_export_keys;
 	crypto_class->get_signature_protocol = gpg_get_signature_protocol;
 	crypto_class->get_encryption_protocol = gpg_get_encryption_protocol;
 	crypto_class->get_key_exchange_protocol = gpg_get_key_exchange_protocol;
+	crypto_class->get_retrieve_session_key = gpg_get_retrieve_session_key;
+	crypto_class->set_retrieve_session_key = gpg_set_retrieve_session_key;
 }
 
 static void
 g_mime_gpg_context_init (GMimeGpgContext *ctx, GMimeGpgContextClass *klass)
 {
-	GMimeCryptoContext *crypto = (GMimeCryptoContext *) ctx;
-	
+	ctx->retrieve_session_key = FALSE;
 	ctx->auto_key_retrieve = FALSE;
 	ctx->always_trust = FALSE;
 	ctx->use_agent = FALSE;
@@ -267,6 +281,7 @@ gpg_get_key_exchange_protocol (GMimeCryptoContext *ctx)
 	return "application/pgp-keys";
 }
 
+#ifdef ENABLE_CRYPTOGRAPHY
 enum _GpgCtxMode {
 	GPG_CTX_MODE_SIGN,
 	GPG_CTX_MODE_VERIFY,
@@ -292,7 +307,10 @@ struct _GpgCtx {
 	int stdout_fd;
 	int stderr_fd;
 	int status_fd;
-	int secret_fd;  /* used for sign/decrypt/verify */
+	int secret_fd;  /* used for exactly one of:
+			 * (a) sending a password to gpg when signing or encrypting
+			 * (b) sending a detatched signature to gpg when verifying
+			 */
 	
 	/* status-fd buffer */
 	char *statusbuf;
@@ -311,6 +329,7 @@ struct _GpgCtx {
 	GMimeCertificateList *encrypted_to;  /* full list of encrypted-to recipients */
 	GMimeSignatureList *signatures;
 	GMimeSignature *signature;
+	char *session_key;
 	
 	int exit_status;
 	
@@ -325,8 +344,10 @@ struct _GpgCtx {
 	unsigned int armor:1;
 	unsigned int need_passwd:1;
 	unsigned int bad_passwds:2;
+	unsigned int decrypt_okay:1;
+	unsigned int override_session_key:1;
 	
-	unsigned int padding:20;
+	unsigned int padding:18;
 };
 
 static struct _GpgCtx *
@@ -340,6 +361,7 @@ gpg_ctx_new (GMimeGpgContext *ctx)
 	gpg->mode = GPG_CTX_MODE_SIGN;
 	gpg->ctx = ctx;
 	gpg->userid_hint = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
+	gpg->decrypt_okay = FALSE;
 	gpg->complete = FALSE;
 	gpg->seen_eof1 = TRUE;
 	gpg->seen_eof2 = FALSE;
@@ -352,6 +374,7 @@ gpg_ctx_new (GMimeGpgContext *ctx)
 	gpg->recipients = NULL;
 	gpg->cipher = GMIME_CIPHER_ALGO_DEFAULT;
 	gpg->digest = GMIME_DIGEST_ALGO_DEFAULT;
+	gpg->override_session_key = FALSE;
 	gpg->always_trust = FALSE;
 	gpg->use_agent = FALSE;
 	gpg->armor = FALSE;
@@ -371,6 +394,7 @@ gpg_ctx_new (GMimeGpgContext *ctx)
 	gpg->need_id = NULL;
 	
 	gpg->encrypted_to = NULL;
+	gpg->session_key = NULL;
 	gpg->signatures = NULL;
 	gpg->signature = NULL;
 	
@@ -380,7 +404,7 @@ gpg_ctx_new (GMimeGpgContext *ctx)
 	
 	stream = g_mime_stream_mem_new ();
 	gpg->diag = GMIME_STREAM_MEM (stream)->buffer;
-	charset = g_mime_locale_charset ();
+	charset = g_mime_charset_iconv_name (g_mime_locale_charset ());
 	if (g_ascii_strcasecmp (charset, "UTF-8") != 0) {
 		GMimeStream *fstream;
 		GMimeFilter *filter;
@@ -408,7 +432,17 @@ static void
 gpg_ctx_set_mode (struct _GpgCtx *gpg, enum _GpgCtxMode mode)
 {
 	gpg->mode = mode;
-	gpg->need_passwd = ((gpg->mode == GPG_CTX_MODE_SIGN) || (gpg->mode == GPG_CTX_MODE_DECRYPT));
+	
+	switch (gpg->mode) {
+	case GPG_CTX_MODE_SIGN_ENCRYPT:
+	case GPG_CTX_MODE_DECRYPT:
+	case GPG_CTX_MODE_SIGN:
+		gpg->need_passwd = TRUE;
+		break;
+	default:
+		gpg->need_passwd = FALSE;
+		break;
+	}
 }
 
 static void
@@ -541,6 +575,11 @@ gpg_ctx_free (struct _GpgCtx *gpg)
 	if (gpg->encrypted_to)
 		g_object_unref (gpg->encrypted_to);
 	
+	if (gpg->session_key) {
+		memset (gpg->session_key, 0, strlen (gpg->session_key));
+		g_free (gpg->session_key);
+	}
+	
 	if (gpg->signatures)
 		g_object_unref (gpg->signatures);
 	
@@ -576,8 +615,16 @@ gpg_digest_str (GMimeDigestAlgo digest)
 	}
 }
 
+static const char *
+filename (const char *path)
+{
+	const char *slash = strrchr (path, '/');
+	
+	return slash != NULL ? slash + 1 : path;
+}
+
 static char **
-gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, int secret_fd, char ***strv)
+gpg_ctx_get_argv (struct _GpgCtx *gpg, const char *path, int status_fd, int secret_fd, char ***strv)
 {
 	const char *digest_str;
 	char **argv, *buf;
@@ -588,7 +635,7 @@ gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, int secret_fd, char ***str
 	*strv = g_new (char *, 3);
 	
 	args = g_ptr_array_new ();
-	g_ptr_array_add (args, "gpg");
+	g_ptr_array_add (args, (char *) filename (path));
 	
 	g_ptr_array_add (args, "--verbose");
 	g_ptr_array_add (args, "--no-secmem-warning");
@@ -600,7 +647,6 @@ gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, int secret_fd, char ***str
                    interactive --command-fd option to send it the
                    user's password */
 		g_ptr_array_add (args, "--batch");
-		g_ptr_array_add (args, "--yes");
 	}
 	
 	g_ptr_array_add (args, "--charset=UTF-8");
@@ -608,7 +654,7 @@ gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, int secret_fd, char ***str
 	(*strv)[v++] = buf = g_strdup_printf ("--status-fd=%d", status_fd);
 	g_ptr_array_add (args, buf);
 	
-	if (gpg->need_passwd) {
+	if (gpg->need_passwd && !gpg->override_session_key) {
 		(*strv)[v++] = buf = g_strdup_printf ("--command-fd=%d", secret_fd);
 		g_ptr_array_add (args, buf);
 	}
@@ -685,6 +731,14 @@ gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, int secret_fd, char ***str
 		if (gpg->use_agent)
 			g_ptr_array_add (args, "--use-agent");
 		
+		if (gpg->ctx->retrieve_session_key)
+			g_ptr_array_add (args, "--show-session-key");
+		
+		if (gpg->override_session_key) {
+			(*strv)[v++] = buf = g_strdup_printf ("--override-session-key-fd=%d", secret_fd);
+			g_ptr_array_add (args, buf);
+		}
+		
 		g_ptr_array_add (args, "--decrypt");
 		g_ptr_array_add (args, "--output");
 		g_ptr_array_add (args, "-");
@@ -718,22 +772,26 @@ gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, int secret_fd, char ***str
 }
 
 static int
-gpg_ctx_op_start (struct _GpgCtx *gpg)
+gpg_ctx_op_start (struct _GpgCtx *gpg, const char *path)
 {
 	int i, maxfd, errnosave, fds[10];
 	char **argv, **strv = NULL;
 	int flags;
 	
-	for (i = 0; i < 10; i++)
+	maxfd = G_N_ELEMENTS (fds);
+	for (i = 0; i < maxfd; i++)
 		fds[i] = -1;
 	
-	maxfd = (gpg->need_passwd || gpg->sigstream) ? 10 : 8;
+	/* don't create the command-fd if we don't need it */
+	if (!(gpg->need_passwd || gpg->sigstream || gpg->override_session_key))
+		maxfd -= 2;
+	
 	for (i = 0; i < maxfd; i += 2) {
 		if (pipe (fds + i) == -1)
 			goto exception;
 	}
 	
-	argv = gpg_ctx_get_argv (gpg, fds[7], fds[8], &strv);
+	argv = gpg_ctx_get_argv (gpg, path, fds[7], fds[8], &strv);
 	
 	if (!(gpg->pid = fork ())) {
 		/* child process */
@@ -815,7 +873,7 @@ gpg_ctx_op_start (struct _GpgCtx *gpg)
 }
 
 static char *
-next_token (char *in, char **token)
+next_token (char *in, gboolean secret, char **token)
 {
 	char *start, *inptr = in;
 	
@@ -832,9 +890,12 @@ next_token (char *in, char **token)
 	while (*inptr && *inptr != ' ' && *inptr != '\n')
 		inptr++;
 	
-	if (token)
+	if (token != NULL)
 		*token = g_strndup (start, (size_t) (inptr - start));
 	
+	if (secret)
+		memset (start, '*', (size_t) (inptr - start));
+	
 	return inptr;
 }
 
@@ -861,7 +922,7 @@ gpg_ctx_add_signature (struct _GpgCtx *gpg, GMimeSignatureStatus status, char *i
 	g_object_unref (sig);
 	
 	/* get the key id of the signer */
-	info = next_token (info, &sig->cert->keyid);
+	info = next_token (info, FALSE, &sig->cert->keyid);
 	
 	/* the rest of the string is the signer's name */
 	sig->cert->name = g_strdup (info);
@@ -901,19 +962,23 @@ gpg_ctx_parse_signer_info (struct _GpgCtx *gpg, char *status)
 		g_object_unref (sig);
 		
 		/* get the key id of the signer */
-		status = next_token (status, &sig->cert->keyid);
+		status = next_token (status, FALSE, &sig->cert->keyid);
 		
 		/* the second token is the public-key algorithm id */
 		sig->cert->pubkey_algo = strtoul (status, &inend, 10);
-		if (inend == status || *inend != ' ')
+		if (inend == status || *inend != ' ') {
+			sig->cert->pubkey_algo = 0;
 			return;
+		}
 		
 		status = inend + 1;
 		
 		/* the third token is the digest algorithm id */
 		sig->cert->digest_algo = strtoul (status, &inend, 10);
-		if (inend == status || *inend != ' ')
+		if (inend == status || *inend != ' ') {
+			sig->cert->digest_algo = 0;
 			return;
+		}
 		
 		status = inend + 1;
 		
@@ -949,10 +1014,10 @@ gpg_ctx_parse_signer_info (struct _GpgCtx *gpg, char *status)
 		status += 9;
 		
 		/* the first token is the fingerprint */
-		status = next_token (status, &sig->cert->fingerprint);
+		status = next_token (status, FALSE, &sig->cert->fingerprint);
 		
 		/* the second token is the date the stream was signed YYYY-MM-DD */
-		status = next_token (status, NULL);
+		status = next_token (status, FALSE, NULL);
 		
 		/* the third token is the signature creation date (or 0 for unknown?) */
 		sig->created = strtoul (status, &inend, 10);
@@ -982,19 +1047,23 @@ gpg_ctx_parse_signer_info (struct _GpgCtx *gpg, char *status)
 		status = inend + 1;
 		
 		/* the sixth token is a reserved numeric value (ignore for now) */
-		status = next_token (status, NULL);
+		status = next_token (status, FALSE, NULL);
 		
 		/* the seventh token is the public-key algorithm id */
 		sig->cert->pubkey_algo = strtoul (status, &inend, 10);
-		if (inend == status || *inend != ' ')
+		if (inend == status || *inend != ' ') {
+			sig->cert->pubkey_algo = 0;
 			return;
+		}
 		
 		status = inend + 1;
 		
 		/* the eighth token is the digest algorithm id */
 		sig->cert->digest_algo = strtoul (status, &inend, 10);
-		if (inend == status || *inend != ' ')
+		if (inend == status || *inend != ' ') {
+			sig->cert->digest_algo = 0;
 			return;
+		}
 		
 		status = inend + 1;
 		
@@ -1026,6 +1095,29 @@ gpg_ctx_parse_signer_info (struct _GpgCtx *gpg, char *status)
 	}
 }
 
+/* write the session_key to the secret file descriptor and close
+   it.  Returns 0 on success. */
+static int
+gpg_ctx_write_session_key (struct _GpgCtx *gpg, const char *session_key)
+{
+	size_t len = strlen (session_key);
+	ssize_t w, nwritten = 0;
+	
+	do {
+		do {
+			w = write (gpg->secret_fd, session_key + nwritten, len - nwritten);
+		} while (w == -1 && (errno == EINTR || errno == EAGAIN));
+		
+		if (w > 0)
+			nwritten += w;
+	} while (nwritten < len && w != -1);
+	
+	close (gpg->secret_fd);
+	gpg->secret_fd = -1;
+	
+	return (w == -1);
+}
+
 static int
 gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 {
@@ -1076,7 +1168,7 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 		
 		status += 12;
 		
-		status = next_token (status, &hint);
+		status = next_token (status, FALSE, &hint);
 		if (!hint) {
 			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
 					     _("Failed to parse gpg userid hint."));
@@ -1100,7 +1192,7 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 		
 		status += 16;
 		
-		status = next_token (status, &userid);
+		status = next_token (status, FALSE, &userid);
 		if (!userid) {
 			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
 					     _("Failed to parse gpg passphrase request."));
@@ -1114,7 +1206,7 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 		
 		status += 20;
 		
-		status = next_token (status, &userid);
+		status = next_token (status, FALSE, &userid);
 		if (!userid) {
 			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
 					     _("Failed to parse gpg passphrase request."));
@@ -1151,7 +1243,7 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 			prompt = g_strdup_printf (_("You need a passphrase to unlock the key for\n"
 						    "user: \"%s\""), name);
 		} else {
-			next_token (status, &prompt);
+			next_token (status, FALSE, &prompt);
 			g_set_error (err, GMIME_ERROR, GMIME_ERROR_GENERAL,
 				     _("Unexpected request from GnuPG for `%s'"), prompt);
 			g_free (prompt);
@@ -1235,10 +1327,10 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 			status += 12;
 			
 			/* skip the next single-char token ("D" for detached) */
-			status = next_token (status, NULL);
+			status = next_token (status, FALSE, NULL);
 			
 			/* skip the public-key algorithm id token */
-			status = next_token (status, NULL);
+			status = next_token (status, FALSE, NULL);
 			
 			/* this token is the digest algorithm used */
 			gpg->digest = strtoul (status, NULL, 10);
@@ -1267,15 +1359,18 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 				
 				/* first token is the mdc algorithm (or 0 if not used) */
 				gpg->digest = strtoul (status, &inend, 10);
-				if (inend == status || *inend != ' ')
-					return;
+				if (inend == status || *inend != ' ') {
+					gpg->digest = 0;
+					break;
+				}
 				
 				status = inend + 1;
 				
 				/* second token is the cipher algorithm */
 				gpg->cipher = strtoul (status, &inend, 10);
 			} else if (!strncmp (status, "DECRYPTION_OKAY", 15)) {
-				/* nothing to do... but good to know gpg decrypted the data successfully */
+				/* decryption succeeded */
+				gpg->decrypt_okay = TRUE;
 			} else if (!strncmp (status, "DECRYPTION_FAILED", 17)) {
 				/* nothing to do... but we know gpg failed to decrypt :-( */
 			} else if (!strncmp (status, "END_DECRYPTION", 14)) {
@@ -1291,13 +1386,17 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 				status += 7;
 				
 				/* first token is the recipient's keyid */
-				status = next_token (status, &cert->keyid);
+				status = next_token (status, FALSE, &cert->keyid);
 				
 				/* second token is the recipient's pubkey algo */
 				cert->pubkey_algo = strtoul (status, &inend, 10);
-				if (inend == status || *inend != ' ')
-					return;
+				if (inend == status || *inend != ' ') {
+					cert->pubkey_algo = 0;
+					g_object_unref (cert);
+					break;
+				}
 				
+				g_object_unref (cert);
 				status = inend + 1;
 				
 				/* third token is a dummy value which is always '0' */
@@ -1305,6 +1404,15 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 				/* nothing to do... we'll grab the MDC used in DECRYPTION_INFO */
 			} else if (!strncmp (status, "BADMDC", 6)) {
 				/* nothing to do, this will only be sent after DECRYPTION_FAILED */
+			} else if (!strncmp (status, "SESSION_KEY", 11)) {
+				if (gpg->session_key) {
+					memset (gpg->session_key, 0, strlen (gpg->session_key));
+					g_free (gpg->session_key);
+				}
+
+				status += 11;
+				
+				status = next_token (status, TRUE, &gpg->session_key);
 			} else {
 				gpg_ctx_parse_signer_info (gpg, status);
 			}
@@ -1542,7 +1650,7 @@ gpg_ctx_op_step (struct _GpgCtx *gpg, GError **err)
 		char buffer[4096];
 		ssize_t nread;
 		
-		d(printf ("reading from gpg's status-fd...\n"));
+		d(printf ("reading gpg's status-fd...\n"));
 		
 		do {
 			nread = read (gpg->status_fd, buffer, sizeof (buffer));
@@ -1792,12 +1900,13 @@ gpg_ctx_op_wait (struct _GpgCtx *gpg)
 	else
 		return -1;
 }
-
+#endif /* ENABLE_CRYPTOGRAPHY */
 
 static int
 gpg_sign (GMimeCryptoContext *context, const char *userid, GMimeDigestAlgo digest,
 	  GMimeStream *istream, GMimeStream *ostream, GError **err)
 {
+#ifdef ENABLE_CRYPTOGRAPHY
 	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
 	struct _GpgCtx *gpg;
 	
@@ -1810,7 +1919,7 @@ gpg_sign (GMimeCryptoContext *context, const char *userid, GMimeDigestAlgo diges
 	gpg_ctx_set_istream (gpg, istream);
 	gpg_ctx_set_ostream (gpg, ostream);
 	
-	if (gpg_ctx_op_start (gpg) == -1) {
+	if (gpg_ctx_op_start (gpg, ctx->path) == -1) {
 		g_set_error (err, GMIME_ERROR, errno,
 			     _("Failed to execute gpg: %s"),
 			     errno ? g_strerror (errno) : _("Unknown"));
@@ -1848,6 +1957,11 @@ gpg_sign (GMimeCryptoContext *context, const char *userid, GMimeDigestAlgo diges
 	gpg_ctx_free (gpg);
 	
 	return digest;
+#else
+	g_set_error (err, GMIME_ERROR, GMIME_ERROR_NOT_SUPPORTED, _("PGP support is not enabled in this build"));
+	
+	return -1;
+#endif /* ENABLE_CRYPTOGRAPHY */
 }
 
 
@@ -1856,11 +1970,10 @@ gpg_verify (GMimeCryptoContext *context, GMimeDigestAlgo digest,
 	    GMimeStream *istream, GMimeStream *sigstream,
 	    GError **err)
 {
+#ifdef ENABLE_CRYPTOGRAPHY
 	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
 	GMimeSignatureList *signatures;
-	const char *diagnostics;
 	struct _GpgCtx *gpg;
-	gboolean valid;
 	
 	gpg = gpg_ctx_new (ctx);
 	gpg_ctx_set_mode (gpg, GPG_CTX_MODE_VERIFY);
@@ -1868,7 +1981,7 @@ gpg_verify (GMimeCryptoContext *context, GMimeDigestAlgo digest,
 	gpg_ctx_set_istream (gpg, istream);
 	gpg_ctx_set_digest (gpg, digest);
 	
-	if (gpg_ctx_op_start (gpg) == -1) {
+	if (gpg_ctx_op_start (gpg, ctx->path) == -1) {
 		g_set_error (err, GMIME_ERROR, errno,
 			     _("Failed to execute gpg: %s"),
 			     errno ? g_strerror (errno) : _("Unknown"));
@@ -1885,7 +1998,8 @@ gpg_verify (GMimeCryptoContext *context, GMimeDigestAlgo digest,
 		}
 	}
 	
-	if (gpg_ctx_op_wait (gpg) != 0) {
+	/* Only set the GError if we got no signature information from gpg */
+	if (gpg_ctx_op_wait (gpg) != 0 && !gpg->signatures) {
 		const char *diagnostics;
 		int save;
 		
@@ -1904,6 +2018,11 @@ gpg_verify (GMimeCryptoContext *context, GMimeDigestAlgo digest,
 	gpg_ctx_free (gpg);
 	
 	return signatures;
+#else
+	g_set_error (err, GMIME_ERROR, GMIME_ERROR_NOT_SUPPORTED, _("PGP support is not enabled in this build"));
+	
+	return NULL;
+#endif /* ENABLE_CRYPTOGRAPHY */
 }
 
 
@@ -1912,6 +2031,7 @@ gpg_encrypt (GMimeCryptoContext *context, gboolean sign, const char *userid,
 	     GMimeDigestAlgo digest, GPtrArray *recipients, GMimeStream *istream,
 	     GMimeStream *ostream, GError **err)
 {
+#ifdef ENABLE_CRYPTOGRAPHY
 	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
 	struct _GpgCtx *gpg;
 	guint i;
@@ -1934,7 +2054,7 @@ gpg_encrypt (GMimeCryptoContext *context, gboolean sign, const char *userid,
 	for (i = 0; i < recipients->len; i++)
 		gpg_ctx_add_recipient (gpg, recipients->pdata[i]);
 	
-	if (gpg_ctx_op_start (gpg) == -1) {
+	if (gpg_ctx_op_start (gpg, ctx->path) == -1) {
 		g_set_error (err, GMIME_ERROR, errno,
 			     _("Failed to execute gpg: %s"),
 			     errno ? g_strerror (errno) : _("Unknown"));
@@ -1969,6 +2089,11 @@ gpg_encrypt (GMimeCryptoContext *context, gboolean sign, const char *userid,
 	gpg_ctx_free (gpg);
 	
 	return 0;
+#else
+	g_set_error (err, GMIME_ERROR, GMIME_ERROR_NOT_SUPPORTED, _("PGP support is not enabled in this build"));
+	
+	return -1;
+#endif /* ENABLE_CRYPTOGRAPHY */
 }
 
 
@@ -1976,6 +2101,15 @@ static GMimeDecryptResult *
 gpg_decrypt (GMimeCryptoContext *context, GMimeStream *istream,
 	     GMimeStream *ostream, GError **err)
 {
+	return gpg_decrypt_session (context, NULL, istream, ostream, err);
+}
+
+static GMimeDecryptResult *
+gpg_decrypt_session (GMimeCryptoContext *context, const char *session_key,
+		     GMimeStream *istream, GMimeStream *ostream,
+		     GError **err)
+{
+#ifdef ENABLE_CRYPTOGRAPHY
 	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
 	GMimeDecryptResult *result;
 	const char *diagnostics;
@@ -1988,7 +2122,10 @@ gpg_decrypt (GMimeCryptoContext *context, GMimeStream *istream,
 	gpg_ctx_set_istream (gpg, istream);
 	gpg_ctx_set_ostream (gpg, ostream);
 	
-	if (gpg_ctx_op_start (gpg) == -1) {
+	if (session_key)
+		gpg->override_session_key = TRUE;
+	
+	if (gpg_ctx_op_start (gpg, ctx->path) == -1) {
 		g_set_error (err, GMIME_ERROR, errno,
 			     _("Failed to execute gpg: %s"),
 			     errno ? g_strerror (errno) : _("Unknown"));
@@ -1997,6 +2134,15 @@ gpg_decrypt (GMimeCryptoContext *context, GMimeStream *istream,
 		return NULL;
 	}
 	
+	if (session_key && gpg_ctx_write_session_key (gpg, session_key)) {
+		g_set_error (err, GMIME_ERROR, errno,
+			     _("Failed to pass session key to gpg: %s"),
+			     errno ? g_strerror (errno) : _("Unknown"));
+		gpg_ctx_free (gpg);
+		
+		return NULL;
+	}
+	
 	while (!gpg_ctx_op_complete (gpg)) {
 		if (gpg_ctx_op_step (gpg, err) == -1) {
 			gpg_ctx_op_cancel (gpg);
@@ -2006,7 +2152,7 @@ gpg_decrypt (GMimeCryptoContext *context, GMimeStream *istream,
 		}
 	}
 	
-	if (gpg_ctx_op_wait (gpg) != 0) {
+	if (gpg_ctx_op_wait (gpg) != 0 && !gpg->decrypt_okay) {
 		save = errno;
 		diagnostics = gpg_ctx_get_diagnostics (gpg);
 		errno = save;
@@ -2020,19 +2166,27 @@ gpg_decrypt (GMimeCryptoContext *context, GMimeStream *istream,
 	result = g_mime_decrypt_result_new ();
 	result->recipients = gpg->encrypted_to;
 	result->signatures = gpg->signatures;
+	result->session_key = gpg->session_key;
 	result->cipher = gpg->cipher;
 	result->mdc = gpg->digest;
 	gpg->encrypted_to = NULL;
 	gpg->signatures = NULL;
+	gpg->session_key = NULL;
 	
 	gpg_ctx_free (gpg);
 	
 	return result;
+#else
+	g_set_error (err, GMIME_ERROR, GMIME_ERROR_NOT_SUPPORTED, _("PGP support is not enabled in this build"));
+	
+	return NULL;
+#endif /* ENABLE_CRYPTOGRAPHY */
 }
 
 static int
 gpg_import_keys (GMimeCryptoContext *context, GMimeStream *istream, GError **err)
 {
+#ifdef ENABLE_CRYPTOGRAPHY
 	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
 	struct _GpgCtx *gpg;
 	
@@ -2040,7 +2194,7 @@ gpg_import_keys (GMimeCryptoContext *context, GMimeStream *istream, GError **err
 	gpg_ctx_set_mode (gpg, GPG_CTX_MODE_IMPORT);
 	gpg_ctx_set_istream (gpg, istream);
 	
-	if (gpg_ctx_op_start (gpg) == -1) {
+	if (gpg_ctx_op_start (gpg, ctx->path) == -1) {
 		g_set_error (err, GMIME_ERROR, errno,
 			     _("Failed to execute gpg: %s"),
 			     errno ? g_strerror (errno) : _("Unknown"));
@@ -2075,11 +2229,17 @@ gpg_import_keys (GMimeCryptoContext *context, GMimeStream *istream, GError **err
 	gpg_ctx_free (gpg);
 	
 	return 0;
+#else
+	g_set_error (err, GMIME_ERROR, GMIME_ERROR_NOT_SUPPORTED, _("PGP support is not enabled in this build"));
+	
+	return -1;
+#endif /* ENABLE_CRYPTOGRAPHY */
 }
 
 static int
 gpg_export_keys (GMimeCryptoContext *context, GPtrArray *keys, GMimeStream *ostream, GError **err)
 {
+#ifdef ENABLE_CRYPTOGRAPHY
 	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
 	struct _GpgCtx *gpg;
 	guint i;
@@ -2093,7 +2253,7 @@ gpg_export_keys (GMimeCryptoContext *context, GPtrArray *keys, GMimeStream *ostr
 		gpg_ctx_add_recipient (gpg, keys->pdata[i]);
 	}
 	
-	if (gpg_ctx_op_start (gpg) == -1) {
+	if (gpg_ctx_op_start (gpg, ctx->path) == -1) {
 		g_set_error (err, GMIME_ERROR, errno,
 			     _("Failed to execute gpg: %s"),
 			     errno ? g_strerror (errno) : _("Unknown"));
@@ -2128,33 +2288,93 @@ gpg_export_keys (GMimeCryptoContext *context, GPtrArray *keys, GMimeStream *ostr
 	gpg_ctx_free (gpg);
 	
 	return 0;
+#else
+	g_set_error (err, GMIME_ERROR, GMIME_ERROR_NOT_SUPPORTED, _("PGP support is not enabled in this build"));
+	
+	return -1;
+#endif /* ENABLE_CRYPTOGRAPHY */
 }
 
+int
+_g_mime_get_gpg_version (const char *path)
+{
+	const char vheader[] = "gpg (GnuPG) ";
+	int v, n = 0, version = 0;
+	const char *inptr;
+	char buffer[128];
+	char *command;
+	FILE *gpg;
+	
+	g_return_val_if_fail (path != NULL, -1);
+	
+	command = g_strdup_printf ("%s --version", path);
+	gpg = popen (command, "r");
+	g_free (command);
+	
+	if (gpg == NULL)
+		return -1;
+	
+	inptr = fgets (buffer, 128, gpg);
+	pclose (gpg);
+	
+	if (strncmp (inptr, vheader, sizeof (vheader) - 1) != 0)
+		return -1;
+	
+	inptr += sizeof (vheader) - 1;
+	while (*inptr >= '0' && *inptr <= '9' && n < 4) {
+		v = 0;
+		
+		while (*inptr >= '0' && *inptr <= '9' && (v < 25 || (v == 25 && *inptr < '6'))) {
+			v = (v * 10) + (*inptr - '0');
+			inptr++;
+		}
+		
+		version = (version << 8) + v;
+		n++;
+		
+		if (*inptr != '.')
+			break;
+		
+		inptr++;
+	}
+	
+	if (n == 0)
+		return -1;
+	
+	if (n < 4)
+		version = version << ((4 - n) * 8);
+	
+	return version;
+}
 
 /**
  * g_mime_gpg_context_new:
  * @request_passwd: a #GMimePasswordRequestFunc
- * @path: path to gpg binary
+ * @path: path to gpg binary or %NULL for the default
  *
  * Creates a new gpg crypto context object.
  *
- * Returns: a new gpg crypto context object.
+ * Returns: (transfer full): a new gpg crypto context object.
  **/
 GMimeCryptoContext *
 g_mime_gpg_context_new (GMimePasswordRequestFunc request_passwd, const char *path)
 {
+#ifdef ENABLE_CRYPTOGRAPHY
 	GMimeCryptoContext *crypto;
 	GMimeGpgContext *ctx;
 	
-	g_return_val_if_fail (path != NULL, NULL);
-	
 	ctx = g_object_newv (GMIME_TYPE_GPG_CONTEXT, 0, NULL);
-	ctx->path = g_strdup (path);
+	ctx->path = g_strdup (path ? path : "gpg");
+	
+	ctx->version = _g_mime_get_gpg_version (ctx->path);
 	
 	crypto = (GMimeCryptoContext *) ctx;
 	crypto->request_passwd = request_passwd;
 	
 	return crypto;
+#else
+	return NULL;
+#endif /* ENABLE_CRYPTOGRAPHY */
 }
 
 
@@ -2162,9 +2382,11 @@ g_mime_gpg_context_new (GMimePasswordRequestFunc request_passwd, const char *pat
  * g_mime_gpg_context_get_auto_key_retrieve:
  * @ctx: a #GMimeGpgContext
  *
- * Gets the @auto_key_retrieve flag on the gpg context.
+ * Gets whether or not gpg should auto-retrieve keys from a keyserver
+ * when verifying signatures.
  *
- * Returns: the @auto_key_retrieve flag on the gpg context.
+ * Returns: %TRUE if gpg should auto-retrieve keys from a keyserver or
+ * %FALSE otherwise.
  **/
 gboolean
 g_mime_gpg_context_get_auto_key_retrieve (GMimeGpgContext *ctx)
@@ -2178,10 +2400,10 @@ g_mime_gpg_context_get_auto_key_retrieve (GMimeGpgContext *ctx)
 /**
  * g_mime_gpg_context_set_auto_key_retrieve:
  * @ctx: a #GMimeGpgContext
- * @auto_key_retrieve: auto-retrieve keys from a keys server
+ * @auto_key_retrieve: %TRUE if gpg should auto-retrieve keys from a keys server
  *
- * Sets the @auto_key_retrieve flag on the gpg context which is used
- * for signature verification.
+ * Sets whether or not gpg should auto-retrieve keys from a keyserver
+ * when verifying signatures.
  **/
 void
 g_mime_gpg_context_set_auto_key_retrieve (GMimeGpgContext *ctx, gboolean auto_key_retrieve)
@@ -2196,9 +2418,10 @@ g_mime_gpg_context_set_auto_key_retrieve (GMimeGpgContext *ctx, gboolean auto_ke
  * g_mime_gpg_context_get_always_trust:
  * @ctx: a #GMimeGpgContext
  *
- * Gets the always_trust flag on the gpg context.
+ * Gets whther or not gpg should always trust keys when encrypting.
  *
- * Returns: the always_trust flag on the gpg context.
+ * Returns: %TRUE if gpg should always trust keys when encrypting or
+ * %FALSE otherwise.
  **/
 gboolean
 g_mime_gpg_context_get_always_trust (GMimeGpgContext *ctx)
@@ -2212,10 +2435,9 @@ g_mime_gpg_context_get_always_trust (GMimeGpgContext *ctx)
 /**
  * g_mime_gpg_context_set_always_trust:
  * @ctx: a #GMimeGpgContext
- * @always_trust: always trust flag
+ * @always_trust: %TRUE if gpg should always trust keys when encrypting
  *
- * Sets the @always_trust flag on the gpg context which is used for
- * encryption.
+ * Sets whether or not gpg should always trust keys when encrypting.
  **/
 void
 g_mime_gpg_context_set_always_trust (GMimeGpgContext *ctx, gboolean always_trust)
@@ -2230,10 +2452,11 @@ g_mime_gpg_context_set_always_trust (GMimeGpgContext *ctx, gboolean always_trust
  * g_mime_gpg_context_get_use_agent:
  * @ctx: a #GMimeGpgContext
  *
- * Gets the use_agent flag on the gpg context.
+ * Gets whether or not gpg should attempt to use the gpg-agent when
+ * requesting credentials.
  *
- * Returns: the use_agent flag on the gpg context, which indicates
- * that GnuPG should attempt to use gpg-agent for credentials.
+ * Returns: %TRUE if the gpg-agent should be used when requesting
+ * credentials or %FALSE otherwise.
  **/
 gboolean
 g_mime_gpg_context_get_use_agent (GMimeGpgContext *ctx)
@@ -2247,10 +2470,10 @@ g_mime_gpg_context_get_use_agent (GMimeGpgContext *ctx)
 /**
  * g_mime_gpg_context_set_use_agent:
  * @ctx: a #GMimeGpgContext
- * @use_agent: always trust flag
+ * @use_agent: use agent flag
  *
- * Sets the @use_agent flag on the gpg context, which indicates that
- * GnuPG should attempt to use gpg-agent for credentials.
+ * Sets whether or not gpg should attempt to use the gpg-agent when
+ * requesting credentials.
  **/
 void
 g_mime_gpg_context_set_use_agent (GMimeGpgContext *ctx, gboolean use_agent)
@@ -2259,3 +2482,31 @@ g_mime_gpg_context_set_use_agent (GMimeGpgContext *ctx, gboolean use_agent)
 	
 	ctx->use_agent = use_agent;
 }
+
+static gboolean
+gpg_get_retrieve_session_key (GMimeCryptoContext *context)
+{
+	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
+	
+	g_return_val_if_fail (GMIME_IS_GPG_CONTEXT (ctx), FALSE);
+	
+	return ctx->retrieve_session_key;
+}
+
+
+static int
+gpg_set_retrieve_session_key (GMimeCryptoContext *context, gboolean retrieve_session_key,
+			      GError **err)
+{
+	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
+	
+	if (!GMIME_IS_GPG_CONTEXT (ctx)) {
+		g_set_error (err, GMIME_ERROR, GMIME_ERROR_NOT_SUPPORTED,
+			     "Not a GMimeGpgContext, can't set retrieve_session_key");
+		return -1;
+	}
+	
+	ctx->retrieve_session_key = retrieve_session_key;
+	
+	return 0;
+}
