@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2010 Jeffrey Stedfast
+ *  Copyright (C) 2000-2014 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -36,7 +36,7 @@
  * SECTION: gmime-message-partial
  * @title: GMimeMessagePartial
  * @short_description: Partial MIME parts
- * @se_also:
+ * @see_also:
  *
  * A #GMimeMessagePartial represents the message/partial MIME part.
  **/
@@ -153,20 +153,12 @@ message_partial_set_header (GMimeObject *object, const char *header, const char
 static const char *
 message_partial_get_header (GMimeObject *object, const char *header)
 {
-	/* RFC 1864 states that you cannot set a Content-MD5 on a message part */
-	if (!g_ascii_strcasecmp ("Content-MD5", header))
-		return NULL;
-	
 	return GMIME_OBJECT_CLASS (parent_class)->get_header (object, header);
 }
 
 static gboolean
 message_partial_remove_header (GMimeObject *object, const char *header)
 {
-	/* RFC 1864 states that you cannot set a Content-MD5 on a message part */
-	if (!g_ascii_strcasecmp ("Content-MD5", header))
-		return FALSE;
-	
 	return GMIME_OBJECT_CLASS (parent_class)->remove_header (object, header);
 }
 
@@ -306,7 +298,8 @@ partial_compare (const void *v1, const void *v2)
  * Reconstructs the GMimeMessage from the given message/partial parts
  * in @partials.
  *
- * Returns: a GMimeMessage object on success or %NULL on fail.
+ * Returns: (transfer full): a GMimeMessage object on success or %NULL
+ * on fail.
  **/
 GMimeMessage *
 g_mime_message_partial_reconstruct_message (GMimeMessagePartial **partials, size_t num)
@@ -411,8 +404,8 @@ message_partial_message_new (GMimeMessage *base)
  * @max_size bytes or fewer. @nparts is set to the number of
  * #GMimeMessagePartial objects created.
  *
- * Returns: an array of #GMimeMessage objects and sets @nparts to th
- * number of messages returned or %NULL on fail.
+ * Returns: (transfer full): an array of #GMimeMessage objects and
+ * sets @nparts to the number of messages returned or %NULL on fail.
  **/
 GMimeMessage **
 g_mime_message_partial_split_message (GMimeMessage *message, size_t max_size, size_t *nparts)
@@ -423,7 +416,7 @@ g_mime_message_partial_split_message (GMimeMessage *message, size_t max_size, si
 	GMimeDataWrapper *wrapper;
 	const unsigned char *buf;
 	GPtrArray *parts;
-	size_t len, end;
+	gint64 len, end;
 	const char *id;
 	gint64 start;
 	guint i;
@@ -443,7 +436,7 @@ g_mime_message_partial_split_message (GMimeMessage *message, size_t max_size, si
 	len = g_mime_stream_length (stream);
 	
 	/* optimization */
-	if (len <= max_size) {
+	if (len <= (gint64) max_size) {
 		g_object_unref (stream);
 		g_object_ref (message);
 		
@@ -462,7 +455,7 @@ g_mime_message_partial_split_message (GMimeMessage *message, size_t max_size, si
 		/* Preferably, we'd split on whole-lines if we can,
 		 * but if that's not possible, split on max size */
 		if ((end = MIN (len, start + max_size)) < len) {
-			register size_t ebx; /* end boundary */
+			register gint64 ebx; /* end boundary */
 			
 			ebx = end;
 			while (ebx > (start + 1) && ebx[buf] != '\n')
