@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2010 Jeffrey Stedfast
+ *  Copyright (C) 2000-2014 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -37,7 +37,13 @@
 #include "gmime-error.h"
 #include "gmime-part.h"
 
+#ifdef ENABLE_DEBUG
+#define d(x) x
+#else
 #define d(x)
+#endif
+
+#define _(x) x
 
 
 /**
@@ -58,14 +64,7 @@ static void g_mime_multipart_signed_init (GMimeMultipartSigned *mps, GMimeMultip
 static void g_mime_multipart_signed_finalize (GObject *object);
 
 /* GMimeObject class methods */
-static void multipart_signed_prepend_header (GMimeObject *object, const char *header, const char *value);
-static void multipart_signed_append_header (GMimeObject *object, const char *header, const char *value);
-static void multipart_signed_set_header (GMimeObject *object, const char *header, const char *value);
-static const char *multipart_signed_get_header (GMimeObject *object, const char *header);
-static gboolean multipart_signed_remove_header (GMimeObject *object, const char *header);
-static void multipart_signed_set_content_type (GMimeObject *object, GMimeContentType *content_type);
-static char *multipart_signed_get_headers (GMimeObject *object);
-static ssize_t multipart_signed_write_to_stream (GMimeObject *object, GMimeStream *stream);
+static void multipart_signed_encode (GMimeObject *object, GMimeEncodingConstraint constraint);
 
 
 static GMimeMultipartClass *parent_class = NULL;
@@ -106,14 +105,7 @@ g_mime_multipart_signed_class_init (GMimeMultipartSignedClass *klass)
 	
 	gobject_class->finalize = g_mime_multipart_signed_finalize;
 	
-	object_class->prepend_header = multipart_signed_prepend_header;
-	object_class->append_header = multipart_signed_append_header;
-	object_class->remove_header = multipart_signed_remove_header;
-	object_class->set_header = multipart_signed_set_header;
-	object_class->get_header = multipart_signed_get_header;
-	object_class->set_content_type = multipart_signed_set_content_type;
-	object_class->get_headers = multipart_signed_get_headers;
-	object_class->write_to_stream = multipart_signed_write_to_stream;
+	object_class->encode = multipart_signed_encode;
 }
 
 static void
@@ -129,51 +121,10 @@ g_mime_multipart_signed_finalize (GObject *object)
 }
 
 static void
-multipart_signed_prepend_header (GMimeObject *object, const char *header, const char *value)
+multipart_signed_encode (GMimeObject *object, GMimeEncodingConstraint constraint)
 {
-	GMIME_OBJECT_CLASS (parent_class)->prepend_header (object, header, value);
-}
-
-static void
-multipart_signed_append_header (GMimeObject *object, const char *header, const char *value)
-{
-	GMIME_OBJECT_CLASS (parent_class)->append_header (object, header, value);
-}
-
-static void
-multipart_signed_set_header (GMimeObject *object, const char *header, const char *value)
-{
-	GMIME_OBJECT_CLASS (parent_class)->set_header (object, header, value);
-}
-
-static const char *
-multipart_signed_get_header (GMimeObject *object, const char *header)
-{
-	return GMIME_OBJECT_CLASS (parent_class)->get_header (object, header);
-}
-
-static gboolean
-multipart_signed_remove_header (GMimeObject *object, const char *header)
-{
-	return GMIME_OBJECT_CLASS (parent_class)->remove_header (object, header);
-}
-
-static void
-multipart_signed_set_content_type (GMimeObject *object, GMimeContentType *content_type)
-{
-	GMIME_OBJECT_CLASS (parent_class)->set_content_type (object, content_type);
-}
-
-static char *
-multipart_signed_get_headers (GMimeObject *object)
-{
-	return GMIME_OBJECT_CLASS (parent_class)->get_headers (object);
-}
-
-static ssize_t
-multipart_signed_write_to_stream (GMimeObject *object, GMimeStream *stream)
-{
-	return GMIME_OBJECT_CLASS (parent_class)->write_to_stream (object, stream);
+	/* Do NOT encode subparts of a multipart/signed */
+	return;
 }
 
 
@@ -247,13 +198,13 @@ sign_prepare (GMimeObject *mime_part)
  * g_mime_multipart_signed_sign:
  * @mps: multipart/signed object
  * @content: MIME part to sign
- * @ctx: encryption cipher context
+ * @ctx: encryption crypto context
  * @userid: user id to sign with
- * @hash: preferred digest algorithm
+ * @digest: preferred digest algorithm
  * @err: exception
  *
  * Attempts to sign the @content MIME part with @userid's private key
- * using the @ctx signing context with the @hash algorithm. If
+ * using the @ctx signing context with the @digest algorithm. If
  * successful, the signed #GMimeObject is set as the signed part of
  * the multipart/signed object @mps.
  *
@@ -263,8 +214,8 @@ sign_prepare (GMimeObject *mime_part)
  **/
 int
 g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
-			      GMimeCipherContext *ctx, const char *userid,
-			      GMimeCipherHash hash, GError **err)
+			      GMimeCryptoContext *ctx, const char *userid,
+			      GMimeDigestAlgo digest, GError **err)
 {
 	GMimeStream *stream, *filtered, *sigstream;
 	GMimeContentType *content_type;
@@ -272,14 +223,19 @@ g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
 	GMimePart *signature;
 	GMimeFilter *filter;
 	GMimeParser *parser;
+	const char *protocol;
 	const char *micalg;
 	int rv;
 	
 	g_return_val_if_fail (GMIME_IS_MULTIPART_SIGNED (mps), -1);
-	g_return_val_if_fail (GMIME_IS_CIPHER_CONTEXT (ctx), -1);
-	g_return_val_if_fail (ctx->sign_protocol != NULL, -1);
+	g_return_val_if_fail (GMIME_IS_CRYPTO_CONTEXT (ctx), -1);
 	g_return_val_if_fail (GMIME_IS_OBJECT (content), -1);
 	
+	if (!(protocol = g_mime_crypto_context_get_signature_protocol (ctx))) {
+		g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_NOT_SUPPORTED, _("Signing not supported."));
+		return -1;
+	}
+	
 	/* Prepare all the parts for signing... */
 	sign_prepare (content);
 	
@@ -312,7 +268,7 @@ g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
 	sigstream = g_mime_stream_mem_new ();
 	
 	/* sign the content stream */
-	if ((rv = g_mime_cipher_context_sign (ctx, userid, hash, filtered, sigstream, err)) == -1) {
+	if ((rv = g_mime_crypto_context_sign (ctx, userid, digest, filtered, sigstream, err)) == -1) {
 		g_object_unref (sigstream);
 		g_object_unref (filtered);
 		g_object_unref (stream);
@@ -325,8 +281,8 @@ g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
 	
 	/* set the multipart/signed protocol and micalg */
 	content_type = g_mime_object_get_content_type (GMIME_OBJECT (mps));
-	g_mime_content_type_set_parameter (content_type, "protocol", ctx->sign_protocol);
-	micalg = g_strdup (g_mime_cipher_context_hash_name (ctx, (GMimeCipherHash) rv));
+	g_mime_content_type_set_parameter (content_type, "protocol", protocol);
+	micalg = g_strdup (g_mime_crypto_context_digest_name (ctx, (GMimeDigestAlgo) rv));
 	g_mime_content_type_set_parameter (content_type, "micalg", micalg);
 	g_mime_multipart_set_boundary (GMIME_MULTIPART (mps), NULL);
 	
@@ -337,7 +293,7 @@ g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
 	g_object_unref (parser);
 	
 	/* construct the signature part */
-	content_type = g_mime_content_type_new_from_string (ctx->sign_protocol);
+	content_type = g_mime_content_type_new_from_string (protocol);
 	signature = g_mime_part_new_with_type (content_type->type, content_type->subtype);
 	g_object_unref (content_type);
 	
@@ -348,9 +304,9 @@ g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
 	g_object_unref (wrapper);
 	
 	/* FIXME: temporary hack, this info should probably be set in
-	 * the CipherContext class - maybe ::sign can take/output a
+	 * the CryptoContext class - maybe ::sign can take/output a
 	 * GMimePart instead. */
-	if (!g_ascii_strcasecmp (ctx->sign_protocol, "application/pkcs7-signature")) {
+	if (!g_ascii_strcasecmp (protocol, "application/pkcs7-signature")) {
 		g_mime_part_set_content_encoding (signature, GMIME_CONTENT_ENCODING_BASE64);
 		g_mime_part_set_filename (signature, "smime.p7m");
 	}
@@ -365,58 +321,99 @@ g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
 	return 0;
 }
 
+static gboolean
+check_protocol_supported (const char *protocol, const char *supported)
+{
+	const char *subtype;
+	char *xsupported;
+	gboolean rv;
+	
+	if (!supported)
+		return FALSE;
+	
+	if (!g_ascii_strcasecmp (protocol, supported))
+		return TRUE;
+	
+	if (!(subtype = strrchr (supported, '/')))
+		return FALSE;
+	
+	subtype++;
+	
+	/* If the subtype already begins with "x-", then there's
+	 * nothing else to check. */
+	if (!g_ascii_strncasecmp (subtype, "x-", 2))
+		return FALSE;
+	
+	/* Check if the "x-" version of the subtype matches the
+	 * protocol. For example, if the supported protocol is
+	 * "application/pkcs7-signature", then we also want to
+	 * match "application/x-pkcs7-signature". */
+	xsupported = g_strdup_printf ("%.*sx-%s", (int) (subtype - supported), supported, subtype);
+	rv = !g_ascii_strcasecmp (protocol, xsupported);
+	g_free (xsupported);
+	
+	return rv;
+}
+
 
 /**
  * g_mime_multipart_signed_verify:
  * @mps: multipart/signed object
- * @ctx: encryption cipher context
+ * @ctx: encryption crypto context
  * @err: exception
  *
  * Attempts to verify the signed MIME part contained within the
- * multipart/signed object @mps using the @ctx cipher context.
+ * multipart/signed object @mps using the @ctx crypto context.
  *
- * Returns: a new #GMimeSignatureValidity object on success or %NULL
- * on fail. If the signing fails, an exception will be set on @err to
- * provide information as to why the failure occured.
+ * Returns: (transfer full): a new #GMimeSignatureList object on
+ * success or %NULL on fail. If the verification fails, an exception
+ * will be set on @err to provide information as to why the failure
+ * occured.
  **/
-GMimeSignatureValidity *
-g_mime_multipart_signed_verify (GMimeMultipartSigned *mps, GMimeCipherContext *ctx,
+GMimeSignatureList *
+g_mime_multipart_signed_verify (GMimeMultipartSigned *mps, GMimeCryptoContext *ctx,
 				GError **err)
 {
+	const char *supported, *protocol, *micalg;
 	GMimeObject *content, *signature;
-	GMimeDataWrapper *wrapper;
-	GMimeStream *filtered_stream;
-	GMimeFilter *crlf_filter;
 	GMimeStream *stream, *sigstream;
-	const char *protocol, *micalg;
-	GMimeSignatureValidity *valid;
-	GMimeCipherHash hash;
+	GMimeSignatureList *signatures;
+	GMimeStream *filtered_stream;
+	GMimeDataWrapper *wrapper;
+	GMimeFilter *crlf_filter;
+	GMimeDigestAlgo digest;
 	char *content_type;
 	
 	g_return_val_if_fail (GMIME_IS_MULTIPART_SIGNED (mps), NULL);
-	g_return_val_if_fail (GMIME_IS_CIPHER_CONTEXT (ctx), NULL);
-	g_return_val_if_fail (ctx->sign_protocol != NULL, NULL);
+	g_return_val_if_fail (GMIME_IS_CRYPTO_CONTEXT (ctx), NULL);
 	
 	if (g_mime_multipart_get_count ((GMimeMultipart *) mps) < 2) {
-		g_set_error (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR, "%s",
-			     "Cannot verify multipart/signed part due to missing subparts.");
+		g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
+				     _("Cannot verify multipart/signed part due to missing subparts."));
 		return NULL;
 	}
 	
 	protocol = g_mime_object_get_content_type_parameter (GMIME_OBJECT (mps), "protocol");
 	micalg = g_mime_object_get_content_type_parameter (GMIME_OBJECT (mps), "micalg");
 	
+	supported = g_mime_crypto_context_get_signature_protocol (ctx);
+	
 	if (protocol) {
-		/* make sure the protocol matches the cipher sign protocol */
-		if (g_ascii_strcasecmp (ctx->sign_protocol, protocol) != 0) {
-			g_set_error (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
-				     "Cannot verify multipart/signed part: unsupported signature protocol '%s'.",
+		/* make sure the protocol matches the crypto sign protocol */
+		if (!check_protocol_supported (protocol, supported)) {
+			g_set_error (err, GMIME_ERROR, GMIME_ERROR_PROTOCOL_ERROR,
+				     _("Cannot verify multipart/signed part: unsupported signature protocol '%s'."),
 				     protocol);
 			return NULL;
 		}
-	} else {
+	} else if (supported != NULL) {
 		/* *shrug* - I guess just go on as if they match? */
-		protocol = ctx->sign_protocol;
+		protocol = supported;
+	} else {
+		g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PROTOCOL_ERROR,
+				     _("Cannot verify multipart/signed part: unspecified signature protocol."));
+		
+		return NULL;
 	}
 	
 	signature = g_mime_multipart_get_part (GMIME_MULTIPART (mps), GMIME_MULTIPART_SIGNED_SIGNATURE);
@@ -424,8 +421,8 @@ g_mime_multipart_signed_verify (GMimeMultipartSigned *mps, GMimeCipherContext *c
 	/* make sure the protocol matches the signature content-type */
 	content_type = g_mime_content_type_to_string (signature->content_type);
 	if (g_ascii_strcasecmp (content_type, protocol) != 0) {
-		g_set_error (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR, "%s",
-			     "Cannot verify multipart/signed part: signature content-type does not match protocol.");
+		g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
+				     _("Cannot verify multipart/signed part: signature content-type does not match protocol."));
 		g_free (content_type);
 		
 		return NULL;
@@ -460,18 +457,20 @@ g_mime_multipart_signed_verify (GMimeMultipartSigned *mps, GMimeCipherContext *c
 		g_mime_data_wrapper_write_to_stream (wrapper, sigstream);
 	} else {
 		sigstream = g_mime_data_wrapper_get_stream (wrapper);
+		g_object_ref (sigstream);
 	}
 	
 	g_mime_stream_reset (sigstream);
 	
 	/* verify the signature */
-	hash = g_mime_cipher_context_hash_id (ctx, micalg);
-	valid = g_mime_cipher_context_verify (ctx, hash, stream, sigstream, err);
+	digest = g_mime_crypto_context_digest_id (ctx, micalg);
+	signatures = g_mime_crypto_context_verify (ctx, digest, stream, sigstream, err);
 	
 	d(printf ("attempted to verify:\n----- BEGIN SIGNED PART -----\n%.*s----- END SIGNED PART -----\n",
 		  (int) GMIME_STREAM_MEM (stream)->buffer->len, GMIME_STREAM_MEM (stream)->buffer->data));
 	
+	g_object_unref (sigstream);
 	g_object_unref (stream);
 	
-	return valid;
+	return signatures;
 }
