@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2009 Jeffrey Stedfast
+ *  Copyright (C) 2000-2014 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -27,9 +27,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <string.h>
-#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-#endif
 #include <fcntl.h>
 #include <errno.h>
 #include <time.h>
@@ -57,14 +55,8 @@ static void g_mime_multipart_init (GMimeMultipart *multipart, GMimeMultipartClas
 static void g_mime_multipart_finalize (GObject *object);
 
 /* GMimeObject class methods */
-static void multipart_prepend_header (GMimeObject *object, const char *header, const char *value);
-static void multipart_append_header (GMimeObject *object, const char *header, const char *value);
-static void multipart_set_header (GMimeObject *object, const char *header, const char *value);
-static const char *multipart_get_header (GMimeObject *object, const char *header);
-static gboolean multipart_remove_header (GMimeObject *object, const char *header);
-static void multipart_set_content_type (GMimeObject *object, GMimeContentType *content_type);
-static char *multipart_get_headers (GMimeObject *object);
 static ssize_t multipart_write_to_stream (GMimeObject *object, GMimeStream *stream);
+static void multipart_encode (GMimeObject *object, GMimeEncodingConstraint constraint);
 
 /* GMimeMultipart class methods */
 static void multipart_clear (GMimeMultipart *multipart);
@@ -118,14 +110,8 @@ g_mime_multipart_class_init (GMimeMultipartClass *klass)
 	
 	gobject_class->finalize = g_mime_multipart_finalize;
 	
-	object_class->prepend_header = multipart_prepend_header;
-	object_class->append_header = multipart_append_header;
-	object_class->remove_header = multipart_remove_header;
-	object_class->set_header = multipart_set_header;
-	object_class->get_header = multipart_get_header;
-	object_class->set_content_type = multipart_set_content_type;
-	object_class->get_headers = multipart_get_headers;
 	object_class->write_to_stream = multipart_write_to_stream;
+	object_class->encode = multipart_encode;
 	
 	klass->add = multipart_add;
 	klass->clear = multipart_clear;
@@ -165,71 +151,6 @@ g_mime_multipart_finalize (GObject *object)
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
-static void
-multipart_prepend_header (GMimeObject *object, const char *header, const char *value)
-{
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a multipart */
-	if (!g_ascii_strncasecmp ("Content-", header, 8))
-		GMIME_OBJECT_CLASS (parent_class)->prepend_header (object, header, value);
-}
-
-static void
-multipart_append_header (GMimeObject *object, const char *header, const char *value)
-{
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a multipart */
-	if (!g_ascii_strncasecmp ("Content-", header, 8))
-		GMIME_OBJECT_CLASS (parent_class)->append_header (object, header, value);
-}
-
-static void
-multipart_set_header (GMimeObject *object, const char *header, const char *value)
-{
-	/* RFC 1864 states that you cannot set a Content-MD5 on a multipart */
-	if (!g_ascii_strcasecmp ("Content-MD5", header))
-		return;
-	
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a multipart */
-	if (!g_ascii_strncasecmp ("Content-", header, 8))
-		GMIME_OBJECT_CLASS (parent_class)->set_header (object, header, value);
-}
-
-static const char *
-multipart_get_header (GMimeObject *object, const char *header)
-{
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a multipart */
-	if (!g_ascii_strncasecmp ("Content-", header, 8))
-		return GMIME_OBJECT_CLASS (parent_class)->get_header (object, header);
-	else
-		return NULL;
-}
-
-static gboolean
-multipart_remove_header (GMimeObject *object, const char *header)
-{
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a multipart */
-	if (g_ascii_strncasecmp ("Content-", header, 8) != 0)
-		return FALSE;
-	
-	return GMIME_OBJECT_CLASS (parent_class)->remove_header (object, header);
-}
-
-static void
-multipart_set_content_type (GMimeObject *object, GMimeContentType *content_type)
-{
-	GMIME_OBJECT_CLASS (parent_class)->set_content_type (object, content_type);
-}
-
-static char *
-multipart_get_headers (GMimeObject *object)
-{
-	return GMIME_OBJECT_CLASS (parent_class)->get_headers (object);
-}
-
 static ssize_t
 multipart_write_to_stream (GMimeObject *object, GMimeStream *stream)
 {
@@ -304,6 +225,19 @@ multipart_write_to_stream (GMimeObject *object, GMimeStream *stream)
 	return total;
 }
 
+static void
+multipart_encode (GMimeObject *object, GMimeEncodingConstraint constraint)
+{
+	GMimeMultipart *multipart = (GMimeMultipart *) object;
+	GMimeObject *subpart;
+	int i;
+	
+	for (i = 0; i < g_mime_multipart_get_count (multipart); i++) {
+		subpart = g_mime_multipart_get_part (multipart, i);
+		g_mime_object_encode (subpart, constraint);
+	}
+}
+
 
 /**
  * g_mime_multipart_new:
@@ -464,7 +398,7 @@ multipart_add (GMimeMultipart *multipart, GMimeObject *part)
  * @multipart: a #GMimeMultipart object
  * @part: a #GMimeObject
  *
- * Adds a mime part to the multipart.
+ * Appends a mime part to @multipart.
  **/
 void
 g_mime_multipart_add (GMimeMultipart *multipart, GMimeObject *part)
@@ -509,11 +443,10 @@ multipart_insert (GMimeMultipart *multipart, int index, GMimeObject *part)
 /**
  * g_mime_multipart_insert:
  * @multipart: a #GMimeMultipart object
- * @part: mime part
- * @index: position to insert the mime part
+ * @part: a #GMimeObject
+ * @index: the 0-based index to insert the part
  *
- * Inserts the specified mime part into the multipart at the position
- * @index.
+ * Inserts @part into @multipart at the specified @index.
  **/
 void
 g_mime_multipart_insert (GMimeMultipart *multipart, int index, GMimeObject *part)
@@ -541,9 +474,9 @@ multipart_remove (GMimeMultipart *multipart, GMimeObject *part)
 /**
  * g_mime_multipart_remove:
  * @multipart: a #GMimeMultipart object
- * @part: mime part
+ * @part: a #GMimeObject
  *
- * Removes the specified mime part from the multipart.
+ * Removes the specified @part from @multipart.
  *
  * Returns: %TRUE if the part was removed or %FALSE otherwise.
  **/
@@ -576,12 +509,12 @@ multipart_remove_at (GMimeMultipart *multipart, int index)
 /**
  * g_mime_multipart_remove_at:
  * @multipart: a #GMimeMultipart object
- * @index: position of the mime part to remove
+ * @index: the 0-based index of the part to remove
  *
- * Removes the mime part at position @index from the multipart.
+ * Removes the part at the specified @index from @multipart.
  *
- * Returns: the mime part that was removed or %NULL if the part was
- * not contained within the multipart.
+ * Returns: (transfer full): the mime part that was removed or %NULL
+ * if the part was not contained within the multipart.
  **/
 GMimeObject *
 g_mime_multipart_remove_at (GMimeMultipart *multipart, int index)
@@ -593,6 +526,38 @@ g_mime_multipart_remove_at (GMimeMultipart *multipart, int index)
 }
 
 
+/**
+ * g_mime_multipart_replace:
+ * @multipart: a #GMimeMultipart object
+ * @index: the 0-based index of the part to replace
+ * @replacement: a #GMimeObject to use as the replacement
+ *
+ * Replaces the part at the specified @index within @multipart with
+ * @replacement.
+ *
+ * Returns: (transfer full): the part that was replaced or %NULL
+ * if the part was not contained within the multipart.
+ **/
+GMimeObject *
+g_mime_multipart_replace (GMimeMultipart *multipart, int index, GMimeObject *replacement)
+{
+	GMimeObject *replaced;
+	
+	g_return_val_if_fail (GMIME_IS_MULTIPART (multipart), NULL);
+	g_return_val_if_fail (GMIME_IS_OBJECT (replacement), NULL);
+	g_return_val_if_fail (index >= 0, NULL);
+	
+	if ((guint) index >= multipart->children->len)
+		return NULL;
+	
+	replaced = multipart->children->pdata[index];
+	multipart->children->pdata[index] = replacement;
+	g_object_ref (replacement);
+	
+	return replaced;
+}
+
+
 static GMimeObject *
 multipart_get_part (GMimeMultipart *multipart, int index)
 {
@@ -610,11 +575,11 @@ multipart_get_part (GMimeMultipart *multipart, int index)
 /**
  * g_mime_multipart_get_part:
  * @multipart: a #GMimeMultipart object
- * @index: position of the mime part
+ * @index: the 0-based index of the part
  *
- * Gets the mime part at position @index within the multipart.
+ * Gets the part at the specified @index within the multipart.
  *
- * Returns: the mime part at position @index.
+ * Returns: (transfer none): the part at position @index.
  **/
 GMimeObject *
 g_mime_multipart_get_part (GMimeMultipart *multipart, int index)
@@ -643,7 +608,7 @@ multipart_contains (GMimeMultipart *multipart, GMimeObject *part)
 /**
  * g_mime_multipart_contains:
  * @multipart: a #GMimeMultipart object
- * @part: mime part
+ * @part: a #GMimeObject
  *
  * Checks if @part is contained within @multipart.
  *
@@ -677,11 +642,11 @@ multipart_index_of (GMimeMultipart *multipart, GMimeObject *part)
 /**
  * g_mime_multipart_index_of:
  * @multipart: a #GMimeMultipart object
- * @part: mime part
+ * @part: a #GMimeObject
  *
- * Gets the index of @part within @multipart.
+ * Gets the 0-based index of @part within @multipart.
  *
- * Returns: the index of @part within @multipart or %-1 if not found.
+ * Returns: the 0-based index of @part within @multipart or %-1 if not found.
  **/
 int
 g_mime_multipart_index_of (GMimeMultipart *multipart, GMimeObject *part)
@@ -704,9 +669,9 @@ multipart_get_count (GMimeMultipart *multipart)
  * g_mime_multipart_get_count:
  * @multipart: a #GMimeMultipart object
  *
- * Gets the number of mime parts contained within the multipart.
+ * Gets the number of parts contained within @multipart.
  *
- * Returns: the number of mime parts contained within the multipart.
+ * Returns: the number of parts contained within @multipart.
  **/
 int
 g_mime_multipart_get_count (GMimeMultipart *multipart)
@@ -720,7 +685,7 @@ g_mime_multipart_get_count (GMimeMultipart *multipart)
 static void
 read_random_pool (unsigned char *buffer, size_t bytes)
 {
-#ifdef HAVE_UNISTD_H
+#ifdef __unix__
 	size_t nread = 0;
 	ssize_t n;
 	int fd;
@@ -745,8 +710,6 @@ read_random_pool (unsigned char *buffer, size_t bytes)
 #else
 	size_t i;
 	
-	srand ((unsigned int) time (NULL));
-	
 	for (i = 0; i < bytes; i++)
 		buffer[i] = (unsigned char) (rand () % 256);
 #endif
@@ -847,7 +810,8 @@ multipart_foreach (GMimeMultipart *multipart, GMimeObjectForeachFunc callback, g
 /**
  * g_mime_multipart_foreach: 
  * @multipart: a #GMimeMultipart
- * @callback: function to call for each of @multipart's subparts.
+ * @callback: (scope call): function to call for each of @multipart's
+ *   subparts.
  * @user_data: user-supplied callback data
  * 
  * Recursively calls @callback on each of @multipart's subparts.
@@ -870,8 +834,8 @@ g_mime_multipart_foreach (GMimeMultipart *multipart, GMimeObjectForeachFunc call
  * Gets the mime part with the content-id @content_id from the
  * multipart @multipart.
  *
- * Returns: the #GMimeObject whose content-id matches the search string,
- * or %NULL if a match cannot be found.
+ * Returns: (transfer none): the #GMimeObject whose content-id matches
+ * the search string, or %NULL if a match cannot be found.
  **/
 GMimeObject *
 g_mime_multipart_get_subpart_from_content_id (GMimeMultipart *multipart, const char *content_id)
