@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2010 Jeffrey Stedfast
+ *  Copyright (C) 2000-2014 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -93,9 +93,9 @@ static void parser_init (GMimeParser *parser, GMimeStream *stream);
 static void parser_close (GMimeParser *parser);
 
 static GMimeObject *parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type,
-						int *found);
+						gboolean toplevel, int *found);
 static GMimeObject *parser_construct_multipart (GMimeParser *parser, ContentType *content_type,
-						int *found);
+						gboolean toplevel, int *found);
 
 static GObjectClass *parent_class = NULL;
 
@@ -114,6 +114,7 @@ enum {
 	GMIME_PARSER_STATE_ERROR = -1,
 	GMIME_PARSER_STATE_INIT,
 	GMIME_PARSER_STATE_FROM,
+	GMIME_PARSER_STATE_MESSAGE_HEADERS,
 	GMIME_PARSER_STATE_HEADERS,
 	GMIME_PARSER_STATE_HEADERS_END,
 	GMIME_PARSER_STATE_CONTENT,
@@ -126,7 +127,7 @@ struct _GMimeParserPrivate {
 	gint64 offset;
 	
 	/* i/o buffers */
-	char realbuf[SCAN_HEAD + SCAN_BUF + 1];
+	char realbuf[SCAN_HEAD + SCAN_BUF + 4];
 	char *inbuf;
 	char *inptr;
 	char *inend;
@@ -179,6 +180,8 @@ struct _GMimeParserPrivate {
 	BoundaryStack *bounds;
 };
 
+static const char MBOX_BOUNDARY[6] = "From ";
+#define MBOX_BOUNDARY_LEN 5
 
 static void
 parser_push_boundary (GMimeParser *parser, const char *boundary)
@@ -193,10 +196,10 @@ parser_push_boundary (GMimeParser *parser, const char *boundary)
 	s->parent = priv->bounds;
 	priv->bounds = s;
 	
-	if (!strcmp (boundary, "From ")) {
-		s->boundary = g_strdup ("From ");
-		s->boundarylen = 5;
-		s->boundarylenfinal = 5;
+	if (boundary == MBOX_BOUNDARY) {
+		s->boundary = g_strdup (boundary);
+		s->boundarylen = MBOX_BOUNDARY_LEN;
+		s->boundarylenfinal = MBOX_BOUNDARY_LEN;
 	} else {
 		s->boundary = g_strdup_printf ("--%s--", boundary);
 		s->boundarylen = strlen (boundary) + 2;
@@ -511,6 +514,8 @@ g_mime_parser_get_persist_stream (GMimeParser *parser)
  * Note: This attribute only serves as a hint to the @parser. If the
  * underlying stream does not support seeking, then this attribute
  * will be ignored.
+ *
+ * By default, this feature is enabled if the underlying stream is seekable.
  **/
 void
 g_mime_parser_set_persist_stream (GMimeParser *parser, gboolean persist)
@@ -569,6 +574,8 @@ g_mime_parser_get_scan_from (GMimeParser *parser)
  * @scan_from: %TRUE to scan From-lines or %FALSE otherwise
  *
  * Sets whether or not @parser should scan mbox-style From-lines.
+ *
+ * By default, this feature is disabled.
  **/
 void
 g_mime_parser_set_scan_from (GMimeParser *parser, gboolean scan_from)
@@ -609,6 +616,8 @@ g_mime_parser_get_respect_content_length (GMimeParser *parser)
  *
  * Most notably useful when parsing broken Solaris mbox files (See
  * http://www.jwz.org/doc/content-length.html for details).
+ *
+ * By default, this feature is disabled.
  **/
 void
 g_mime_parser_set_respect_content_length (GMimeParser *parser, gboolean respect_content_length)
@@ -825,7 +834,7 @@ parser_step_from (GMimeParser *parser)
 	
  got_from:
 	
-	priv->state = GMIME_PARSER_STATE_HEADERS;
+	priv->state = GMIME_PARSER_STATE_MESSAGE_HEADERS;
 	
 	priv->inptr = inptr;
 	
@@ -865,7 +874,7 @@ nearest_pow (size_t num)
 static inline size_t
 next_alloc_size (size_t n)
 {
-	return ((n + 63) / 64) * 64;
+	return (n + 63) & ~63;
 }
 #endif
 
@@ -956,12 +965,61 @@ header_parse (GMimeParser *parser, HeaderRaw **tail)
 				 header->offset, priv->user_data);
 #elif defined (HAVE_REGEX_H)
 	if (priv->have_regex &&
-	    !regexec (&priv->regex, header->name, 0, NULL, 0))
+	    !regexec (&priv->header_regex, header->name, 0, NULL, 0))
 		priv->header_cb (parser, header->name, header->value,
 				 header->offset, priv->user_data);
 #endif
 }
 
+enum {
+	SUBJECT = 1 << 0,
+	FROM    = 1 << 1,
+	DATE    = 1 << 2,
+	TO      = 1 << 3,
+	CC      = 1 << 4
+};
+
+static gboolean
+has_message_headers (HeaderRaw *headers)
+{
+	unsigned int found = 0;
+	HeaderRaw *header;
+	
+	header = headers;
+	while (header != NULL) {
+		if (!g_ascii_strcasecmp (header->name, "Subject"))
+			found |= SUBJECT;
+		else if (!g_ascii_strcasecmp (header->name, "From"))
+			found |= FROM;
+		else if (!g_ascii_strcasecmp (header->name, "Date"))
+			found |= DATE;
+		else if (!g_ascii_strcasecmp (header->name, "To"))
+			found |= TO;
+		else if (!g_ascii_strcasecmp (header->name, "Cc"))
+			found |= CC;
+		
+		header = header->next;
+	}
+	
+	return found != 0;
+}
+
+static gboolean
+has_content_headers (HeaderRaw *headers)
+{
+	HeaderRaw *header;
+	
+	header = headers;
+	while (header != NULL) {
+		if (!g_ascii_strcasecmp (header->name, "Content-Type"))
+			return TRUE;
+		
+		header = header->next;
+	}
+	
+	return FALSE;
+}
+
 static int
 parser_step_headers (GMimeParser *parser)
 {
@@ -1033,7 +1091,7 @@ parser_step_headers (GMimeParser *parser)
 					
 					/* Note: see optimization comment [1] */
 					*inend = '\n';
-				} else if (*inptr == ':') {
+				} else {
 					valid = FALSE;
 				}
 				
@@ -1042,11 +1100,27 @@ parser_step_headers (GMimeParser *parser)
 					    && !strncmp (start, "From ", 5))
 						goto next_message;
 					
-					if (priv->headers != NULL || *inptr == ':') {
-						/* probably the start of the content,
-						 * a broken mailer didn't terminate the
-						 * headers with an empty line. *sigh* */
-						goto content_start;
+					if (priv->headers != NULL) {
+						if (priv->state == GMIME_PARSER_STATE_MESSAGE_HEADERS) {
+							if (has_message_headers (priv->headers)) {
+								/* probably the start of the content,
+								 * a broken mailer didn't terminate the
+								 * headers with an empty line. *sigh* */
+								goto content_start;
+							}
+						} else if (has_content_headers (priv->headers)) {
+							/* probably the start of the content,
+							 * a broken mailer didn't terminate the
+							 * headers with an empty line. *sigh* */
+							goto content_start;
+						}
+					} else if (priv->state == GMIME_PARSER_STATE_MESSAGE_HEADERS) {
+						/* Be a little more strict when scanning toplevel message
+						 * headers, but remain lenient with From-lines. */
+						if ((inptr - start) != 4 || strncmp (start, "From ", 5) != 0) {
+							priv->state = GMIME_PARSER_STATE_ERROR;
+							return -1;
+						}
 					}
 				}
 			}
@@ -1166,7 +1240,7 @@ content_type_is_type (ContentType *content_type, const char *type, const char *s
 }
 
 static ContentType *
-parser_content_type (GMimeParser *parser)
+parser_content_type (GMimeParser *parser, GMimeContentType *parent)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	ContentType *content_type;
@@ -1176,8 +1250,13 @@ parser_content_type (GMimeParser *parser)
 	
 	if (!(value = header_raw_find (priv->headers, "Content-Type", NULL)) ||
 	    !g_mime_parse_content_type (&value, &content_type->type, &content_type->subtype)) {
-		content_type->type = g_strdup ("text");
-		content_type->subtype = g_strdup ("plain");
+		if (parent != NULL && g_mime_content_type_is_type (parent, "multipart", "digest")) {
+			content_type->type = g_strdup ("message");
+			content_type->subtype = g_strdup ("rfc822");
+		} else {
+			content_type->type = g_strdup ("text");
+			content_type->subtype = g_strdup ("plain");
+		}
 	}
 	
 	content_type->exists = value != NULL;
@@ -1234,13 +1313,14 @@ parser_step (GMimeParser *parser)
 		if (priv->scan_from)
 			priv->state = GMIME_PARSER_STATE_FROM;
 		else
-			priv->state = GMIME_PARSER_STATE_HEADERS;
+			priv->state = GMIME_PARSER_STATE_MESSAGE_HEADERS;
 		break;
 	case GMIME_PARSER_STATE_FROM:
 		priv->message_headers_begin = -1;
 		priv->message_headers_end = -1;
 		parser_step_from (parser);
 		break;
+	case GMIME_PARSER_STATE_MESSAGE_HEADERS:
 	case GMIME_PARSER_STATE_HEADERS:
 		parser_step_headers (parser);
 		
@@ -1284,6 +1364,33 @@ enum {
                          ((scan_from && len >= 5 && !strncmp (start, "From ", 5)) ||  \
 			  (len >= 2 && (start[0] == '-' && start[1] == '-')))
 
+static gboolean
+is_boundary (const char *text, size_t len, const char *boundary, size_t boundary_len)
+{
+	const char *inptr = text + boundary_len;
+	const char *inend = text + len;
+	
+	if (boundary_len > len)
+		return FALSE;
+	
+	/* make sure that the text matches the boundary */
+	if (strncmp (text, boundary, boundary_len) != 0)
+		return FALSE;
+	
+	if (!strncmp (text, "From ", 5))
+		return TRUE;
+	
+	/* the boundary may be optionally followed by linear whitespace */
+	while (inptr < inend) {
+		if (!is_lwsp (*inptr))
+			return FALSE;
+		
+		inptr++;
+	}
+	
+	return TRUE;
+}
+
 static int
 check_boundary (struct _GMimeParserPrivate *priv, const char *start, size_t len)
 {
@@ -1299,19 +1406,14 @@ check_boundary (struct _GMimeParserPrivate *priv, const char *start, size_t len)
 		
 		s = priv->bounds;
 		while (s) {
-			/* we use >= here because From lines are > 5 chars */
 			if (offset >= s->content_end &&
-			    len >= s->boundarylenfinal &&
-			    !strncmp (s->boundary, start,
-				      s->boundarylenfinal)) {
+			    is_boundary (start, len, s->boundary, s->boundarylenfinal)) {
 				d(printf ("found %s\n", s->content_end != -1 && offset >= s->content_end ?
 					  "end of content" : "end boundary"));
 				return FOUND_END_BOUNDARY;
 			}
 			
-			if (len == s->boundarylen &&
-			    !strncmp (s->boundary, start,
-				      s->boundarylen)) {
+			if (is_boundary (start, len, s->boundary, s->boundarylen)) {
 				d(printf ("found boundary\n"));
 				return FOUND_BOUNDARY;
 			}
@@ -1325,6 +1427,22 @@ check_boundary (struct _GMimeParserPrivate *priv, const char *start, size_t len)
 	return FOUND_NOTHING;
 }
 
+static gboolean
+found_immediate_boundary (struct _GMimeParserPrivate *priv, gboolean end)
+{
+	BoundaryStack *s = priv->bounds;
+	size_t boundary_len = end ? s->boundarylenfinal : s->boundarylen;
+	register char *inptr = priv->inptr;
+	char *inend = priv->inend;
+	
+	/* Note: see optimization comment [1] */
+	*inend = '\n';
+	while (*inptr != '\n')
+		inptr++;
+	
+	return is_boundary (priv->inptr, inptr - priv->inptr, s->boundary, boundary_len);
+}
+
 /* Optimization Notes:
  *
  * 1. By making the priv->realbuf char array 1 extra char longer, we
@@ -1343,11 +1461,13 @@ static int
 parser_scan_content (GMimeParser *parser, GByteArray *content, guint *crlf)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
+	char *aligned, *start, *inend;
 	register char *inptr;
-	char *start, *inend;
+	register int *dword;
 	size_t nleft, len;
 	size_t atleast;
 	int found = 0;
+	int mask;
 	
 	d(printf ("scan-content\n"));
 	
@@ -1381,11 +1501,26 @@ parser_scan_content (GMimeParser *parser, GByteArray *content, guint *crlf)
 		priv->midline = FALSE;
 		
 		while (inptr < inend) {
+			aligned = (char *) (((long) (inptr + 3)) & ~3);
 			start = inptr;
+			
 			/* Note: see optimization comment [1] */
-			while (*inptr != '\n')
+			while (inptr < aligned && *inptr != '\n')
 				inptr++;
 			
+			if (inptr == aligned) {
+				dword = (int *) inptr;
+				
+				do {
+					mask = *dword++ ^ 0x0A0A0A0A;
+					mask = ((mask - 0x01010101) & (~mask & 0x80808080));
+				} while (mask == 0);
+				
+				inptr = (char *) (dword - 1);
+				while (*inptr != '\n')
+					inptr++;
+			}
+			
 			len = (size_t) (inptr - start);
 			
 			if (inptr < inend) {
@@ -1489,6 +1624,40 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 	
 	g_assert (priv->state == GMIME_PARSER_STATE_CONTENT);
 	
+	if (priv->bounds != NULL) {
+		/* Check for the possibility of an empty message/rfc822 part. */
+		register char *inptr;
+		size_t atleast;
+		char *inend;
+		
+		/* figure out minimum amount of data we need */
+		atleast = MAX (SCAN_HEAD, MAX_BOUNDARY_LEN (priv->bounds));
+		
+		if (parser_fill (parser, atleast) <= 0) {
+			*found = FOUND_EOS;
+			return;
+		}
+		
+		inptr = priv->inptr;
+		inend = priv->inend;
+		/* Note: see optimization comment [1] */
+		*inend = '\n';
+		
+		while (*inptr != '\n')
+			inptr++;
+		
+		*found = check_boundary (priv, priv->inptr, inptr - priv->inptr);
+		switch (*found) {
+		case FOUND_END_BOUNDARY:
+			/* ignore "From " boundaries, boken mailers tend to include these lines... */
+			if (strncmp (priv->inptr, "From ", 5) != 0)
+				return;
+			break;
+		case FOUND_BOUNDARY:
+			return;
+		}
+	}
+	
 	/* get the headers */
 	priv->state = GMIME_PARSER_STATE_HEADERS;
 	if (parser_step (parser) == GMIME_PARSER_STATE_ERROR) {
@@ -1501,15 +1670,16 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 	message = g_mime_message_new (FALSE);
 	header = priv->headers;
 	while (header) {
-		g_mime_object_append_header ((GMimeObject *) message, header->name, header->value);
+		if (g_ascii_strncasecmp (header->name, "Content-", 8) != 0)
+			g_mime_object_append_header ((GMimeObject *) message, header->name, header->value);
 		header = header->next;
 	}
 	
-	content_type = parser_content_type (parser);
+	content_type = parser_content_type (parser, NULL);
 	if (content_type_is_type (content_type, "multipart", "*"))
-		object = parser_construct_multipart (parser, content_type, found);
+		object = parser_construct_multipart (parser, content_type, TRUE, found);
 	else
-		object = parser_construct_leaf_part (parser, content_type, found);
+		object = parser_construct_leaf_part (parser, content_type, TRUE, found);
 	
 	content_type_destroy (content_type);
 	message->mime_part = object;
@@ -1523,7 +1693,7 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 }
 
 static GMimeObject *
-parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, int *found)
+parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, gboolean toplevel, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	GMimeObject *object;
@@ -1537,14 +1707,15 @@ parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, int
 	if (!content_type->exists) {
 		GMimeContentType *mime_type;
 		
-		mime_type = g_mime_content_type_new ("text", "plain");
+		mime_type = g_mime_content_type_new (content_type->type, content_type->subtype);
 		_g_mime_object_set_content_type (object, mime_type);
 		g_object_unref (mime_type);
 	}
 	
 	header = priv->headers;
 	while (header) {
-		g_mime_object_append_header (object, header->name, header->value);
+		if (!toplevel || !g_ascii_strncasecmp (header->name, "Content-", 8))
+			g_mime_object_append_header (object, header->name, header->value);
 		header = header->next;
 	}
 	
@@ -1634,16 +1805,6 @@ parser_scan_multipart_face (GMimeParser *parser, GMimeMultipart *multipart, gboo
 #define parser_scan_multipart_preface(parser, multipart) parser_scan_multipart_face (parser, multipart, TRUE)
 #define parser_scan_multipart_postface(parser, multipart) parser_scan_multipart_face (parser, multipart, FALSE)
 
-static gboolean
-found_immediate_boundary (struct _GMimeParserPrivate *priv, gboolean end)
-{
-	BoundaryStack *s = priv->bounds;
-	size_t len = end ? s->boundarylenfinal : s->boundarylen;
-	
-	return !strncmp (priv->inptr, s->boundary, len)
-		&& (priv->inptr[len] == '\n' || priv->inptr[len] == '\r');
-}
-
 static int
 parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 {
@@ -1671,11 +1832,11 @@ parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 			break;
 		}
 		
-		content_type = parser_content_type (parser);
+		content_type = parser_content_type (parser, ((GMimeObject *) multipart)->content_type);
 		if (content_type_is_type (content_type, "multipart", "*"))
-			subpart = parser_construct_multipart (parser, content_type, &found);
+			subpart = parser_construct_multipart (parser, content_type, FALSE, &found);
 		else
-			subpart = parser_construct_leaf_part (parser, content_type, &found);
+			subpart = parser_construct_leaf_part (parser, content_type, FALSE, &found);
 		
 		g_mime_multipart_add (multipart, subpart);
 		content_type_destroy (content_type);
@@ -1686,7 +1847,7 @@ parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 }
 
 static GMimeObject *
-parser_construct_multipart (GMimeParser *parser, ContentType *content_type, int *found)
+parser_construct_multipart (GMimeParser *parser, ContentType *content_type, gboolean toplevel, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	GMimeMultipart *multipart;
@@ -1701,7 +1862,8 @@ parser_construct_multipart (GMimeParser *parser, ContentType *content_type, int
 	
 	header = priv->headers;
 	while (header) {
-		g_mime_object_append_header (object, header->name, header->value);
+		if (!toplevel || !g_ascii_strncasecmp (header->name, "Content-", 8))
+			g_mime_object_append_header (object, header->name, header->value);
 		header = header->next;
 	}
 	
@@ -1763,16 +1925,17 @@ parser_construct_part (GMimeParser *parser)
 	int found;
 	
 	/* get the headers */
+	priv->state = GMIME_PARSER_STATE_HEADERS;
 	while (priv->state < GMIME_PARSER_STATE_HEADERS_END) {
 		if (parser_step (parser) == GMIME_PARSER_STATE_ERROR)
 			return NULL;
 	}
 	
-	content_type = parser_content_type (parser);
+	content_type = parser_content_type (parser, NULL);
 	if (content_type_is_type (content_type, "multipart", "*"))
-		object = parser_construct_multipart (parser, content_type, &found);
+		object = parser_construct_multipart (parser, content_type, FALSE, &found);
 	else
-		object = parser_construct_leaf_part (parser, content_type, &found);
+		object = parser_construct_leaf_part (parser, content_type, FALSE, &found);
 	
 	content_type_destroy (content_type);
 	
@@ -1786,7 +1949,8 @@ parser_construct_part (GMimeParser *parser)
  *
  * Constructs a MIME part from @parser.
  *
- * Returns: a MIME part based on @parser or %NULL on fail.
+ * Returns: (transfer full): a MIME part based on @parser or %NULL on
+ * fail.
  **/
 GMimeObject *
 g_mime_parser_construct_part (GMimeParser *parser)
@@ -1811,7 +1975,7 @@ parser_construct_message (GMimeParser *parser)
 	int found;
 	
 	/* scan the from-line if we are parsing an mbox */
-	while (priv->state != GMIME_PARSER_STATE_HEADERS) {
+	while (priv->state != GMIME_PARSER_STATE_MESSAGE_HEADERS) {
 		if (parser_step (parser) == GMIME_PARSER_STATE_ERROR)
 			return NULL;
 	}
@@ -1831,21 +1995,22 @@ parser_construct_message (GMimeParser *parser)
 				content_length = ULONG_MAX;
 		}
 		
-		g_mime_object_append_header ((GMimeObject *) message, header->name, header->value);
+		if (g_ascii_strncasecmp (header->name, "Content-", 8) != 0)
+			g_mime_object_append_header ((GMimeObject *) message, header->name, header->value);
 		header = header->next;
 	}
 	
 	if (priv->scan_from) {
-		parser_push_boundary (parser, "From ");
+		parser_push_boundary (parser, MBOX_BOUNDARY);
 		if (priv->respect_content_length && content_length < ULONG_MAX)
 			priv->bounds->content_end = parser_offset (priv, NULL) + content_length;
 	}
 	
-	content_type = parser_content_type (parser);
+	content_type = parser_content_type (parser, NULL);
 	if (content_type_is_type (content_type, "multipart", "*"))
-		object = parser_construct_multipart (parser, content_type, &found);
+		object = parser_construct_multipart (parser, content_type, TRUE, &found);
 	else
-		object = parser_construct_leaf_part (parser, content_type, &found);
+		object = parser_construct_leaf_part (parser, content_type, TRUE, &found);
 	
 	content_type_destroy (content_type);
 	message->mime_part = object;
@@ -1869,7 +2034,7 @@ parser_construct_message (GMimeParser *parser)
  *
  * Constructs a MIME message from @parser.
  *
- * Returns: a MIME message or %NULL on fail.
+ * Returns: (transfer full): a MIME message or %NULL on fail.
  **/
 GMimeMessage *
 g_mime_parser_construct_message (GMimeParser *parser)
