@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2010 Jeffrey Stedfast
+ *  Copyright (C) 2000-2014 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -34,8 +34,10 @@
 #include "gmime-stream-null.h"
 #include "gmime-stream-filter.h"
 #include "gmime-filter-basic.h"
+#include "gmime-filter-best.h"
 #include "gmime-filter-crlf.h"
 #include "gmime-filter-md5.h"
+#include "gmime-table-private.h"
 
 #define d(x)
 
@@ -59,10 +61,9 @@ static void g_mime_part_finalize (GObject *object);
 static void mime_part_prepend_header (GMimeObject *object, const char *header, const char *value);
 static void mime_part_append_header (GMimeObject *object, const char *header, const char *value);
 static void mime_part_set_header (GMimeObject *object, const char *header, const char *value);
-static const char *mime_part_get_header (GMimeObject *object, const char *header);
 static gboolean mime_part_remove_header (GMimeObject *object, const char *header);
-static char *mime_part_get_headers (GMimeObject *object);
 static ssize_t mime_part_write_to_stream (GMimeObject *object, GMimeStream *stream);
+static void mime_part_encode (GMimeObject *object, GMimeEncodingConstraint constraint);
 
 /* GMimePart class methods */
 static void set_content_object (GMimePart *mime_part, GMimeDataWrapper *content);
@@ -110,9 +111,8 @@ g_mime_part_class_init (GMimePartClass *klass)
 	object_class->append_header = mime_part_append_header;
 	object_class->remove_header = mime_part_remove_header;
 	object_class->set_header = mime_part_set_header;
-	object_class->get_header = mime_part_get_header;
-	object_class->get_headers = mime_part_get_headers;
 	object_class->write_to_stream = mime_part_write_to_stream;
+	object_class->encode = mime_part_encode;
 	
 	klass->set_content_object = set_content_object;
 }
@@ -159,23 +159,41 @@ static const char *content_headers[] = {
 };
 
 
+static void
+copy_atom (const char *src, char *dest, size_t n)
+{
+	register const char *inptr = src;
+	register char *outptr = dest;
+	char *outend = dest + n;
+	
+	while (is_lwsp (*inptr))
+		inptr++;
+	
+	while (is_atom (*inptr) && outptr < outend)
+		*outptr++ = *inptr++;
+	
+	*outptr = '\0';
+}
+
 static gboolean
 process_header (GMimeObject *object, const char *header, const char *value)
 {
 	GMimePart *mime_part = (GMimePart *) object;
-	char *text;
+	char encoding[32];
 	guint i;
 	
+	if (g_ascii_strncasecmp (header, "Content-", 8) != 0)
+		return FALSE;
+	
 	for (i = 0; i < G_N_ELEMENTS (content_headers); i++) {
-		if (!g_ascii_strcasecmp (content_headers[i], header))
+		if (!g_ascii_strcasecmp (content_headers[i] + 8, header + 8))
 			break;
 	}
 	
 	switch (i) {
 	case HEADER_CONTENT_TRANSFER_ENCODING:
-		text = g_mime_strdup_trim (value);
-		mime_part->encoding = g_mime_content_encoding_from_string (text);
-		g_free (text);
+		copy_atom (value, encoding, sizeof (encoding) - 1);
+		mime_part->encoding = g_mime_content_encoding_from_string (encoding);
 		break;
 	case HEADER_CONTENT_DESCRIPTION:
 		/* FIXME: we should decode this */
@@ -200,11 +218,6 @@ process_header (GMimeObject *object, const char *header, const char *value)
 static void
 mime_part_prepend_header (GMimeObject *object, const char *header, const char *value)
 {
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a mime part */
-	if (g_ascii_strncasecmp ("Content-", header, 8) != 0)
-		return;
-	
 	if (!process_header (object, header, value))
 		GMIME_OBJECT_CLASS (parent_class)->prepend_header (object, header, value);
 	else
@@ -214,11 +227,6 @@ mime_part_prepend_header (GMimeObject *object, const char *header, const char *v
 static void
 mime_part_append_header (GMimeObject *object, const char *header, const char *value)
 {
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a mime part */
-	if (g_ascii_strncasecmp ("Content-", header, 8) != 0)
-		return;
-	
 	if (!process_header (object, header, value))
 		GMIME_OBJECT_CLASS (parent_class)->append_header (object, header, value);
 	else
@@ -228,74 +236,48 @@ mime_part_append_header (GMimeObject *object, const char *header, const char *va
 static void
 mime_part_set_header (GMimeObject *object, const char *header, const char *value)
 {
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a mime part */
-	if (g_ascii_strncasecmp ("Content-", header, 8) != 0)
-		return;
-	
 	if (!process_header (object, header, value))
 		GMIME_OBJECT_CLASS (parent_class)->set_header (object, header, value);
 	else
 		g_mime_header_list_set (object->headers, header, value);
 }
 
-static const char *
-mime_part_get_header (GMimeObject *object, const char *header)
-{
-	/* Make sure that the header is a Content-* header, else it
-           doesn't belong on a mime part */
-	if (!g_ascii_strncasecmp ("Content-", header, 8))
-		return GMIME_OBJECT_CLASS (parent_class)->get_header (object, header);
-	else
-		return NULL;
-}
-
 static gboolean
 mime_part_remove_header (GMimeObject *object, const char *header)
 {
 	GMimePart *mime_part = (GMimePart *) object;
 	guint i;
 	
-	/* Make sure that the header is a Content-* header, else it
-	   doesn't belong on a mime part */
-	if (g_ascii_strncasecmp ("Content-", header, 8) != 0)
-		return FALSE;
-	
-	for (i = 0; i < G_N_ELEMENTS (content_headers); i++) {
-		if (!g_ascii_strcasecmp (content_headers[i], header))
+	if (!g_ascii_strncasecmp (header, "Content-", 8)) {
+		for (i = 0; i < G_N_ELEMENTS (content_headers); i++) {
+			if (!g_ascii_strcasecmp (content_headers[i] + 8, header + 8))
+				break;
+		}
+		
+		switch (i) {
+		case HEADER_CONTENT_TRANSFER_ENCODING:
+			mime_part->encoding = GMIME_CONTENT_ENCODING_DEFAULT;
 			break;
-	}
-	
-	switch (i) {
-	case HEADER_CONTENT_TRANSFER_ENCODING:
-		mime_part->encoding = GMIME_CONTENT_ENCODING_DEFAULT;
-		break;
-	case HEADER_CONTENT_DESCRIPTION:
-		/* FIXME: we should decode this */
-		g_free (mime_part->content_description);
-		mime_part->content_description = NULL;
-		break;
-	case HEADER_CONTENT_LOCATION:
-		g_free (mime_part->content_location);
-		mime_part->content_location = NULL;
-		break;
-	case HEADER_CONTENT_MD5:
-		g_free (mime_part->content_md5);
-		mime_part->content_md5 = NULL;
-		break;
-	default:
-		break;
+		case HEADER_CONTENT_DESCRIPTION:
+			g_free (mime_part->content_description);
+			mime_part->content_description = NULL;
+			break;
+		case HEADER_CONTENT_LOCATION:
+			g_free (mime_part->content_location);
+			mime_part->content_location = NULL;
+			break;
+		case HEADER_CONTENT_MD5:
+			g_free (mime_part->content_md5);
+			mime_part->content_md5 = NULL;
+			break;
+		default:
+			break;
+		}
 	}
 	
 	return GMIME_OBJECT_CLASS (parent_class)->remove_header (object, header);
 }
 
-static char *
-mime_part_get_headers (GMimeObject *object)
-{
-	return GMIME_OBJECT_CLASS (parent_class)->get_headers (object);
-}
-
 static ssize_t
 write_content (GMimePart *part, GMimeStream *stream)
 {
@@ -362,6 +344,7 @@ write_content (GMimePart *part, GMimeStream *stream)
 		content_stream = g_mime_data_wrapper_get_stream (part->content);
 		g_mime_stream_reset (content_stream);
 		nwritten = g_mime_stream_write_to_stream (content_stream, stream);
+		g_mime_stream_reset (content_stream);
 		
 		if (nwritten == -1)
 			return -1;
@@ -398,6 +381,54 @@ mime_part_write_to_stream (GMimeObject *object, GMimeStream *stream)
 	return total;
 }
 
+static void
+mime_part_encode (GMimeObject *object, GMimeEncodingConstraint constraint)
+{
+	GMimePart *part = (GMimePart *) object;
+	GMimeContentEncoding encoding;
+	GMimeStream *stream, *null;
+	GMimeFilter *filter;
+	
+	switch (part->encoding) {
+	case GMIME_CONTENT_ENCODING_DEFAULT:
+		/* Unspecified encoding, we need to figure out the
+		 * best encoding no matter what */
+		break;
+	case GMIME_CONTENT_ENCODING_7BIT:
+		/* This encoding is always safe. */
+		return;
+	case GMIME_CONTENT_ENCODING_8BIT:
+		/* This encoding is safe unless the constraint is 7bit. */
+		if (constraint != GMIME_ENCODING_CONSTRAINT_7BIT)
+			return;
+		break;
+	case GMIME_CONTENT_ENCODING_BINARY:
+		/* This encoding is only safe if the constraint is binary. */
+		if (constraint == GMIME_ENCODING_CONSTRAINT_BINARY)
+			return;
+		break;
+	case GMIME_CONTENT_ENCODING_BASE64:
+	case GMIME_CONTENT_ENCODING_QUOTEDPRINTABLE:
+	case GMIME_CONTENT_ENCODING_UUENCODE:
+		/* These encodings are always safe. */
+		return;
+	}
+	
+	filter = g_mime_filter_best_new (GMIME_FILTER_BEST_ENCODING);
+	
+	null = g_mime_stream_null_new ();
+	stream = g_mime_stream_filter_new (null);
+	g_mime_stream_filter_add ((GMimeStreamFilter *) stream, filter);
+	g_object_unref (null);
+	
+	g_mime_object_write_to_stream (object, stream);
+	g_object_unref (stream);
+	
+	encoding = g_mime_filter_best_encoding ((GMimeFilterBest *) filter, constraint);
+	g_mime_part_set_content_encoding (part, encoding);
+	g_object_unref (filter);
+}
+
 
 /**
  * g_mime_part_new:
@@ -480,7 +511,7 @@ g_mime_part_set_content_description (GMimePart *mime_part, const char *descripti
  * Returns: the content description for the specified mime part.
  **/
 const char *
-g_mime_part_get_content_description (const GMimePart *mime_part)
+g_mime_part_get_content_description (GMimePart *mime_part)
 {
 	g_return_val_if_fail (GMIME_IS_PART (mime_part), NULL);
 	
@@ -734,13 +765,78 @@ g_mime_part_get_content_encoding (GMimePart *mime_part)
 }
 
 
+/**
+ * g_mime_part_get_best_content_encoding:
+ * @mime_part: a #GMimePart object
+ * @constraint: a #GMimeEncodingConstraint
+ *
+ * Calculates the most efficient content encoding for the @mime_part
+ * given the @constraint.
+ *
+ * Returns: the best content encoding for the specified mime part.
+ **/
+GMimeContentEncoding
+g_mime_part_get_best_content_encoding (GMimePart *mime_part, GMimeEncodingConstraint constraint)
+{
+	GMimeStream *filtered, *stream;
+	GMimeContentEncoding encoding;
+	GMimeFilterBest *best;
+	GMimeFilter *filter;
+	
+	g_return_val_if_fail (GMIME_IS_PART (mime_part), GMIME_CONTENT_ENCODING_DEFAULT);
+	
+	stream = g_mime_stream_null_new ();
+	filtered = g_mime_stream_filter_new (stream);
+	g_object_unref (stream);
+	
+	filter = g_mime_filter_best_new (GMIME_FILTER_BEST_ENCODING);
+	g_mime_stream_filter_add ((GMimeStreamFilter *) filtered, filter);
+	best = (GMimeFilterBest *) filter;
+	
+	g_mime_data_wrapper_write_to_stream (mime_part->content, filtered);
+	g_mime_stream_flush (filtered);
+	g_object_unref (filtered);
+	
+	encoding = g_mime_filter_best_encoding (best, constraint);
+	g_object_unref (best);
+	
+	return encoding;
+}
+
+
+/**
+ * g_mime_part_is_attachment:
+ * @mime_part: a #GMimePart object
+ *
+ * Determines whether or not the part is an attachment based on the
+ * value of the Content-Disposition header.
+ *
+ * Returns: %TRUE if the part is an attachment, otherwise %FALSE.
+ *
+ * Since: 2.6.21
+ **/
+gboolean
+g_mime_part_is_attachment (GMimePart *mime_part)
+{
+	GMimeContentDisposition *disposition;
+	
+	g_return_val_if_fail (GMIME_IS_PART (mime_part), FALSE);
+	
+	disposition = g_mime_object_get_content_disposition ((GMimeObject *) mime_part);
+	
+	return disposition != NULL && g_mime_content_disposition_is_attachment (disposition);
+}
+
+
 /**
  * g_mime_part_set_filename:
  * @mime_part: a #GMimePart object
- * @filename: the filename of the Mime Part's content
+ * @filename: the file name
  *
  * Sets the "filename" parameter on the Content-Disposition and also sets the
  * "name" parameter on the Content-Type.
+ *
+ * Note: The @filename string should be in UTF-8.
  **/
 void
 g_mime_part_set_filename (GMimePart *mime_part, const char *filename)
@@ -758,15 +854,15 @@ g_mime_part_set_filename (GMimePart *mime_part, const char *filename)
  * g_mime_part_get_filename:
  * @mime_part: a #GMimePart object
  *
- * Gets the filename of the specificed mime part, or %NULL if the mime
- * part does not have the filename or name parameter set.
+ * Gets the filename of the specificed mime part, or %NULL if the
+ * @mime_part does not have the filename or name parameter set.
  *
- * Returns: the filename of the specified MIME Part. It first checks to
- * see if the "filename" parameter was set on the Content-Disposition
- * and if not then checks the "name" parameter in the Content-Type.
+ * Returns: the filename of the specified @mime_part or %NULL if
+ * neither of the parameters is set. If a file name is set, the
+ * returned string will be in UTF-8.
  **/
 const char *
-g_mime_part_get_filename (const GMimePart *mime_part)
+g_mime_part_get_filename (GMimePart *mime_part)
 {
 	GMimeObject *object = (GMimeObject *) mime_part;
 	const char *filename = NULL;
@@ -791,6 +887,7 @@ set_content_object (GMimePart *mime_part, GMimeDataWrapper *content)
 	g_object_ref (content);
 }
 
+
 /**
  * g_mime_part_set_content_object:
  * @mime_part: a #GMimePart object
@@ -817,10 +914,11 @@ g_mime_part_set_content_object (GMimePart *mime_part, GMimeDataWrapper *content)
  * Gets the internal data-wrapper of the specified mime part, or %NULL
  * on error.
  *
- * Returns: the data-wrapper for the mime part's contents.
+ * Returns: (transfer none): the data-wrapper for the mime part's
+ * contents.
  **/
 GMimeDataWrapper *
-g_mime_part_get_content_object (const GMimePart *mime_part)
+g_mime_part_get_content_object (GMimePart *mime_part)
 {
 	g_return_val_if_fail (GMIME_IS_PART (mime_part), NULL);
 	
