@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2010 Jeffrey Stedfast
+ *  Copyright (C) 2000-2014 Jeffrey Stedfast
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -28,7 +28,6 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
-#include <dirent.h>
 #include <fcntl.h>
 #include <time.h>
 
@@ -87,52 +86,72 @@ print_mime_struct (GMimeStream *stream, GMimeObject *part, int depth)
 	}
 }
 
-
-static void
-header_cb (GMimeParser *parser, const char *header, const char *value, gint64 offset, gpointer user_data)
-{
-	GMimeStream *stream = user_data;
-	
-	g_mime_stream_printf (stream, "%lld: %s: %s\n", offset, header, value);
-}
-
 static void
 test_parser (GMimeParser *parser, GMimeStream *mbox, GMimeStream *summary)
 {
-	gint64 headers_begin, headers_end, start, end;
+	gint64 message_begin, message_end, headers_begin, headers_end;
+	InternetAddressList *list;
 	GMimeMessage *message;
-	const char *exev;
+	char *marker, *buf;
+	const char *subject;
+	const char *sender;
+	int tz_offset;
 	int nmsg = 0;
-	char *from;
+	time_t date;
 	
 	while (!g_mime_parser_eos (parser)) {
-		start = g_mime_parser_tell (parser);
+		message_begin = g_mime_parser_tell (parser);
 		if (!(message = g_mime_parser_construct_message (parser)))
 			throw (exception_new ("failed to parse message #%d", nmsg));
 		
-		end = g_mime_parser_tell (parser);
+		message_end = g_mime_parser_tell (parser);
 		
 		headers_begin = g_mime_parser_get_headers_begin (parser);
 		headers_end = g_mime_parser_get_headers_end (parser);
 		
-		g_mime_stream_printf (summary, "message offsets: %lld, %lld\n", start, end);
-		g_mime_stream_printf (summary, "header offsets: %lld, %lld\n",
+		g_mime_stream_printf (summary, "message offsets: %" G_GINT64_FORMAT ", %" G_GINT64_FORMAT "\n",
+				      message_begin, message_end);
+		g_mime_stream_printf (summary, "header offsets: %" G_GINT64_FORMAT ", %" G_GINT64_FORMAT "\n",
 				      headers_begin, headers_end);
 		
-		from = g_mime_parser_get_from (parser);
-		g_mime_stream_printf (summary, "%s\n", from);
-		exev = g_mime_object_get_header ((GMimeObject *) message, "X-Evolution");
-		g_mime_stream_printf (summary, "X-Evolution: %s\n", exev ? exev : "None");
+		marker = g_mime_parser_get_from (parser);
+		g_mime_stream_printf (summary, "%s\n", marker);
+		
+		if ((sender = g_mime_message_get_sender (message)) != NULL) {
+			list = internet_address_list_parse_string (sender);
+			buf = internet_address_list_to_string (list, FALSE);
+			g_object_unref (list);
+			
+			g_mime_stream_printf (summary, "From: %s\n", buf);
+			g_free (buf);
+		}
+		
+		if ((list = g_mime_message_get_recipients (message, GMIME_RECIPIENT_TYPE_TO)) != NULL &&
+		    internet_address_list_length (list) > 0) {
+			buf = internet_address_list_to_string (list, FALSE);
+			g_mime_stream_printf (summary, "To: %s\n", buf);
+			g_free (buf);
+		}
+		
+		if (!(subject = g_mime_message_get_subject (message)))
+			subject = "";
+		g_mime_stream_printf (summary, "Subject: %s\n", subject);
+		
+		g_mime_message_get_date (message, &date, &tz_offset);
+		buf = g_mime_utils_header_format_date (date, tz_offset);
+		g_mime_stream_printf (summary, "Date: %s\n", buf);
+		g_free (buf);
+		
 		print_mime_struct (summary, message->mime_part, 0);
 		g_mime_stream_write (summary, "\n", 1);
 		
 		if (mbox) {
-			g_mime_stream_printf (mbox, "%s%s\n", nmsg > 0 ? "\n" : "", from);
+			g_mime_stream_printf (mbox, "%s%s\n", nmsg > 0 ? "\n" : "", marker);
 			g_mime_object_write_to_stream ((GMimeObject *) message, mbox);
 		}
 		
 		g_object_unref (message);
-		g_free (from);
+		g_free (marker);
 		nmsg++;
 	}
 }
@@ -141,7 +160,7 @@ static gboolean
 streams_match (GMimeStream *istream, GMimeStream *ostream)
 {
 	char buf[4096], dbuf[4096], errstr[1024], *bufptr, *bufend, *dbufptr;
-	size_t totalsize, totalread = 0;
+	gint64 len, totalsize, totalread = 0;
 	size_t nread, size;
 	gint64 offset = 0;
 	ssize_t n;
@@ -150,14 +169,14 @@ streams_match (GMimeStream *istream, GMimeStream *ostream)
 	
 	if (istream->bound_end != -1) {
 		totalsize = istream->bound_end - istream->position;
-	} else if ((n = g_mime_stream_length (istream)) == -1) {
+	} else if ((len = g_mime_stream_length (istream)) == -1) {
 		sprintf (errstr, "Error: Unable to get length of original stream\n");
 		goto fail;
-	} else if (n < (istream->position - istream->bound_start)) {
+	} else if (len < (istream->position - istream->bound_start)) {
 		sprintf (errstr, "Error: Overflow on original stream?\n");
 		goto fail;
 	} else {
-		totalsize = n - (istream->position - istream->bound_start);
+		totalsize = len - (istream->position - istream->bound_start);
 	}
 	
 	while (totalread < totalsize) {
@@ -198,7 +217,7 @@ streams_match (GMimeStream *istream, GMimeStream *ostream)
 		}
 		
 		if (bufptr < bufend) {
-			sprintf (errstr, "Error: content does not match at offset %lld\n",
+			sprintf (errstr, "Error: content does not match at offset %" G_GINT64_FORMAT "\n",
 				 offset + (bufptr - buf));
 			/*fprintf (stderr, "-->'%.*s'<--\nvs\n-->'%.*s'<--\n",
 			  bufend - bufptr, bufptr, bufend - bufptr, dbufptr);*/
@@ -238,11 +257,11 @@ int main (int argc, char **argv)
 	char input[256], output[256], *tmp, *p, *q;
 	GMimeStream *istream, *ostream, *mstream, *pstream;
 	GMimeParser *parser;
-	struct dirent *dent;
+	const char *dent;
 	const char *path;
 	struct stat st;
+	GDir *dir;
 	int fd, i;
-	DIR *dir;
 	
 #ifdef ENABLE_MBOX_MATCH
 	if (mkdir ("./tmp", 0755) == -1 && errno != EEXIST)
@@ -272,7 +291,7 @@ int main (int argc, char **argv)
 		*p++ = G_DIR_SEPARATOR;
 		p = g_stpcpy (p, "input");
 		
-		if (!(dir = opendir (input)))
+		if (!(dir = g_dir_open (input, 0, NULL)))
 			goto exit;
 		
 		*p++ = G_DIR_SEPARATOR;
@@ -284,12 +303,12 @@ int main (int argc, char **argv)
 		*q++ = G_DIR_SEPARATOR;
 		*q = '\0';
 		
-		while ((dent = readdir (dir))) {
-			if (!g_str_has_suffix (dent->d_name, ".mbox"))
+		while ((dent = g_dir_read_name (dir))) {
+			if (!g_str_has_suffix (dent, ".mbox"))
 				continue;
 			
-			strcpy (p, dent->d_name);
-			strcpy (q, dent->d_name);
+			strcpy (p, dent);
+			strcpy (q, dent);
 			
 			tmp = NULL;
 			parser = NULL;
@@ -298,27 +317,27 @@ int main (int argc, char **argv)
 			mstream = NULL;
 			pstream = NULL;
 			
-			testsuite_check ("%s", dent->d_name);
+			testsuite_check ("%s", dent);
 			try {
-				if ((fd = open (input, O_RDONLY)) == -1) {
+				if ((fd = open (input, O_RDONLY, 0)) == -1) {
 					throw (exception_new ("could not open `%s': %s",
-							      input, strerror (errno)));
+							      input, g_strerror (errno)));
 				}
 				
 				istream = g_mime_stream_fs_new (fd);
 				
-				if ((fd = open (output, O_RDONLY)) == -1) {
+				if ((fd = open (output, O_RDONLY, 0)) == -1) {
 					throw (exception_new ("could not open `%s': %s",
-							      output, strerror (errno)));
+							      output, g_strerror (errno)));
 				}
 				
 				ostream = g_mime_stream_fs_new (fd);
 				
 #ifdef ENABLE_MBOX_MATCH
-				tmp = g_strdup_printf ("./tmp/%s.XXXXXX", dent->d_name);
+				tmp = g_strdup_printf ("./tmp/%s.XXXXXX", dent);
 				if ((fd = g_mkstemp (tmp)) == -1) {
 					throw (exception_new ("could not open `%s': %s",
-							      tmp, strerror (errno)));
+							      tmp, g_strerror (errno)));
 				}
 				
 				mstream = g_mime_stream_fs_new (fd);
@@ -328,24 +347,23 @@ int main (int argc, char **argv)
 				g_mime_parser_set_persist_stream (parser, TRUE);
 				g_mime_parser_set_scan_from (parser, TRUE);
 				
-				if (strstr (dent->d_name, "content-length") != NULL)
+				if (strstr (dent, "content-length") != NULL)
 					g_mime_parser_set_respect_content_length (parser, TRUE);
 				
 				pstream = g_mime_stream_mem_new ();
-				g_mime_parser_set_header_regex (parser, "^Subject$", header_cb, pstream);
 				test_parser (parser, mstream, pstream);
 				
 #ifdef ENABLE_MBOX_MATCH
 				g_mime_stream_reset (istream);
 				g_mime_stream_reset (mstream);
 				if (!streams_match (istream, mstream))
-					throw (exception_new ("mboxes do not match for `%s'", dent->d_name));
+					throw (exception_new ("mboxes do not match for `%s'", dent));
 #endif
 				
 				g_mime_stream_reset (ostream);
 				g_mime_stream_reset (pstream);
 				if (!streams_match (ostream, pstream))
-					throw (exception_new ("summaries do not match for `%s'", dent->d_name));
+					throw (exception_new ("summaries do not match for `%s'", dent));
 				
 				testsuite_check_passed ();
 				
@@ -354,9 +372,9 @@ int main (int argc, char **argv)
 #endif
 			} catch (ex) {
 				if (parser != NULL)
-					testsuite_check_failed ("%s: %s", dent->d_name, ex->message);
+					testsuite_check_failed ("%s: %s", dent, ex->message);
 				else
-					testsuite_check_warn ("%s: %s", dent->d_name, ex->message);
+					testsuite_check_warn ("%s: %s", dent, ex->message);
 			} finally;
 			
 			if (mstream != NULL)
@@ -377,10 +395,10 @@ int main (int argc, char **argv)
 			g_free (tmp);
 		}
 		
-		closedir (dir);
+		g_dir_close (dir);
 	} else if (S_ISREG (st.st_mode)) {
 		/* manually run test on a single file */
-		if ((fd = open (path, O_RDONLY)) == -1)
+		if ((fd = open (path, O_RDONLY, 0)) == -1)
 			goto exit;
 		
 		istream = g_mime_stream_fs_new (fd);
@@ -402,7 +420,6 @@ int main (int argc, char **argv)
 #endif
 		
 		ostream = g_mime_stream_fs_new (dup (1));
-		g_mime_parser_set_header_regex (parser, "^Subject$", header_cb, ostream);
 		
 		testsuite_check ("user-input mbox: `%s'", path);
 		try {
